#include "AnimatedModelContainer.h"
#include "EchoChamber.h"
#include <assert.h>
#include <QSqlQuery>
#include <QSqlError>

static AnimatedModel *animatedModelContainerHead=0;

AnimatedModel::AnimatedModel()
{
	next=0;
	if (animatedModelContainerHead==0)
	{
		animatedModelContainerHead=this;
	}
	else
	{
		AnimatedModel *first = animatedModelContainerHead;
		while (first->next)
			first=first->next;
		first->next=this;
	}
	lastRowIndex=currentRowIndex=0;
	useAnimationRowIndices=false;
	ClearAnimationRowIndices();
}
QString AnimatedModel::InitDB(QString queryString, QString postfixTickDefinition, QString databaseName)
{
	model.setQuery(queryString, QSqlDatabase::database(databaseName));
	SetTickDefinition(postfixTickDefinition);
	QSqlError error = model.lastError();
	if (error.isValid()==true)
	{
		QString err = error.databaseText();
		return err;
	}
	dbName=databaseName;
	return QString();
}
AnimatedModel::~AnimatedModel()
{
	if (animatedModelContainerHead==this)
	{
		animatedModelContainerHead=animatedModelContainerHead->next;
	}
	else
	{
		AnimatedModel *first = animatedModelContainerHead;
		while (first->next!=this)
			first=first->next;
		first->next=first->next->next;
	}
}
AnimatedModel *AnimatedModel::GetLinkedList(void)
{
	return animatedModelContainerHead;
}
void AnimatedModel::SetTickDefinition(QString postfixTickDefinition)
{
	hasValidQuery=tableRowQuery.SetQuery(postfixTickDefinition);
	if (hasValidQuery)
		hasValidQuery=tableRowQuery.Preparse(&model);
}
bool AnimatedModel::GetDoubleAtIndex(int idx, int columnIndex, double *d)
{
	if (idx >= model.realRowCount() || idx < 0)
		return false;
	QVariant v = model.data(model.index(idx, columnIndex));
	if (v.isNull()==false)
	{
		*d=v.toDouble();
		return true;
	}
	return false;
}
bool AnimatedModel::GetDoubleAtIndex(int idx, QString columnName, double *d)
{
	int columnIndex = EchoChamber::GetColumnIndex(&model, columnName);
	if (columnIndex==-1)
		return false;
	return GetDoubleAtIndex(idx,columnIndex,d);
}
bool AnimatedModel::GetIndexOfValue(int *idxBefore, int *idxAfter, double *valueBefore, double *valueAfter, double valueIn, QString columnName)
{
	double minimumTime, maximumTime;
	int columnIndex = EchoChamber::GetColumnIndex(&model, columnName);
	if (model.GetNumericalColumnConstraints(columnIndex, &minimumTime, &maximumTime))
	{
		if (valueIn < minimumTime || valueIn > maximumTime)
			return false;

		int rowCount = model.realRowCount();
		double proportion = (valueIn-minimumTime) / (maximumTime-minimumTime);
		int lastRow;
		int nextRow;
		double valueAtLastRow, valueAtNextRow;
		QVariant valueVariant;
		lastRow=nextRow=proportion*(rowCount-1);		
		valueVariant=model.data(model.index(nextRow,columnIndex));
		while (valueVariant.isNull() && nextRow>=0)
		{
			nextRow--;
			lastRow--;
			valueVariant=model.data(model.index(nextRow,columnIndex));
		}

		if (nextRow<0)
			return false;

		valueAtLastRow=valueAtNextRow=valueVariant.toDouble();

		if (valueAtNextRow>=valueIn)
		{
			// subtract indices and find next row value
			while (nextRow>=0 && valueAtNextRow>=valueIn)
			{
				if (valueAtNextRow==valueIn)
				{
					*idxBefore=nextRow;
					*valueBefore=valueAtLastRow;
					*idxAfter=AdvanceWhileEqual(nextRow,columnIndex,rowCount,1,valueAtNextRow);
					*valueAfter=valueAtLastRow;
					return true;
				}

				nextRow--;
				lastRow=nextRow;
				nextRow=AdvanceWhileEqual(nextRow,columnIndex,rowCount,-1,valueAtNextRow);
				valueVariant=model.data(model.index(nextRow,columnIndex));
				valueAtLastRow=valueAtNextRow;
				valueAtNextRow=valueVariant.toDouble();
			}
			if (nextRow<0)
			{
				*idxBefore=lastRow;
				*idxAfter=0;
				*valueBefore=valueAtLastRow;
				*valueAfter=valueAtNextRow;
				return true;
			}

			*idxBefore=lastRow;
			*valueBefore=valueAtLastRow;
			*idxAfter=nextRow;
			*valueAfter=valueAtLastRow;
			return true;
		}
		else
		{
			// Add indices and find next row value
			while (nextRow<rowCount && valueAtNextRow<=valueIn)
			{
				if (valueAtNextRow==valueIn)
				{
					*idxBefore=nextRow;
					*valueBefore=valueAtLastRow;
					*idxAfter=AdvanceWhileEqual(nextRow,columnIndex,rowCount,1,valueAtNextRow);
					*valueAfter=valueAtLastRow;
					return true;
				}

				nextRow++;
				lastRow=nextRow;
				nextRow=AdvanceWhileEqual(nextRow,columnIndex,rowCount,1,valueAtNextRow);
				valueVariant=model.data(model.index(nextRow,columnIndex));
				valueAtLastRow=valueAtNextRow;
				valueAtNextRow=valueVariant.toDouble();
			}
			if (nextRow==rowCount)
			{
				*idxBefore=lastRow;
				*idxAfter=rowCount;
				*valueBefore=valueAtLastRow;
				*valueAfter=valueAtNextRow;
				return true;
			}

			*idxBefore=lastRow;
			*valueBefore=valueAtLastRow;
			*idxAfter=nextRow;
			*valueAfter=valueAtLastRow;
			return true;
		}
	}

	return false;
}
int AnimatedModel::AdvanceWhileEqual(int nextRow, int columnIndex, int rowCount, int step, double valueToCheck)
{
	int nextRowToCheck;

	QVariant valueVariant;
	while (1)
	{
		if ((nextRow==0 && step<0) || (nextRow==rowCount-1 && step>0))
			return nextRow;
		nextRowToCheck = nextRow + step;
		valueVariant=model.data(model.index(nextRowToCheck,columnIndex));
		if (valueVariant.isNull()==false)
		{
			if (valueVariant.toDouble()!=valueToCheck)
				return nextRow;
		}
		nextRow=nextRowToCheck;
	}
	return nextRow;
}
void AnimatedModel::GetAnimationRowIndices(int *rowStart, int *rowEnd, int fullTableRowConstraint)
{
	if (UseAnimationRowIndices())
	{
		*rowStart=GetAnimationFromIndex();
		*rowEnd=GetAnimationToIndex()+1;
	}
	else
	{
		*rowStart=0;
		if (fullTableRowConstraint==0)
		{
			*rowEnd=model.realRowCount();
		}
		else
		{
			*rowEnd=model.constrainedRowCount(fullTableRowConstraint);
			if (*rowEnd>fullTableRowConstraint)
				*rowEnd=fullTableRowConstraint;
		}
	}

}
void AnimatedModel::SetAnimationRowIndices(int from, int to, double valueBefore, double valueAfter)
{
	assert(from>=0);
	animationRowIndexFrom=from;
	int realRowCount = model.realRowCount();
	assert(to<realRowCount);
	animationRowIndexTo=to; 
	useAnimationRowIndices=true;
}
void AnimatedModel::ClearAnimationRowIndices(void)
{
	useAnimationRowIndices=false;
}
int AnimatedModel::GetNextRowStep(int currentIndex, bool forward)
{
	if (CanRowStep())
		return tableRowQuery.GetNextRowStepIndexInclusive(GetModel(),currentIndex,forward);
	return -1;
}
void AnimatedModel::AMC_Save(XMLSaver &xmlSaver)
{
	tableRowQuery.Save(xmlSaver);
}
void AnimatedModel::AMC_Load(XMLSaver &xmlSaver)
{
	tableRowQuery.Load(xmlSaver);
	SetTickDefinition(tableRowQuery.GetTickDefinition());
}