#ifndef ECHOCHAMBER_H
#define ECHOCHAMBER_H

#include <QtGui/QMainWindow>
#include <QtGui/QWorkspace>
#include <QTreeWidgetItem>
#include <QSqlQuery>
#include "QSqlQueryModelFetchAll.h"
#include <QList>
#include <QString>
#include <QTimer>
#include <QTime>
#include "qwt_scale_draw.h"
#include "ui_EchoChamber.h"
#include "EchoChamberSettings.h"
#include "ISaveableWindow.h"

#define DATABASE_TYPE "QSQLITEEC"
//#define DATABASE_TYPE "QSQLITE"

class SQLQueryTool;
class TableWindow;
class PieWindow;
class QwtWindow;
class ScatterPlotWindow;
class QtColorComboBox;
class QComboBox;
class AnimatedModel;
class AnimatedModelList;
class ImageViewer;

class ScaleDrawNoScientific : public QwtScaleDraw
{
	QwtText label(double value) const;
};

class EchoChamber : public QMainWindow, public Ui::EchoChamberClass
{
	Q_OBJECT
public:
	EchoChamber(QWidget *parent = 0, Qt::WFlags flags = 0);
	~EchoChamber();

	SQLQueryTool *SpawnSQLQueryTool(QStringList queries, QStringList databaseNames);
	TableWindow *SpawnTableWindow(QStringList queries, QStringList databaseNames);
	PieWindow *SpawnPieWindow(QStringList queries, QStringList databaseNames);
	QwtWindow *SpawnQwtWindow(QStringList queries, QStringList databaseNames ,int initialRenderMode);
	ScatterPlotWindow *SpawnScatterPlotWindow(QStringList queries, QStringList databaseNames, bool startInHeatMapMode);
	ImageViewer *SpawnImageViewer(void);
	const QStringList& GetDatabases(void) const {return databases;}
	void AddToLastQueryLineEdit(const QString& query);
	void FlagNeedsSave(void);

	int SetupXYZTimeComboBoxes(const QSqlQueryModelFetchAll *model, QComboBox *xComboBox, QComboBox *yComboBox, QComboBox *zComboBox, QComboBox *timeComboBox);
	void SetupSymbolMagnitudeComboBox(const QSqlQueryModelFetchAll *model, QComboBox *comboBox);
	void SetupIntensityComboBox(const QSqlQueryModelFetchAll *model, QComboBox *comboBox);
	void SetupRGBComboBox(const QSqlQueryModelFetchAll *model, QComboBox *comboBox, bool *defaultWasSet);
	static int GetColumnIndex(const QSqlQueryModelFetchAll *model, const QString &columnName);
	void GetColumnTypes(
		const QSqlQueryModelFetchAll *model,
		QStringList &allColumnNames,
		QStringList &numericalColumnNames, QList<int> &numericalColumnIndices,
		QStringList &textColumnNames, QList<int> &textColumnIndices,
		int &timestampColumnIndex);
	static QString GetProductVersionName(void) {return "Echo Chamber Beta 1";}
	static int GetLayoutVersionNumber(void) {return 0;}

	void AddDefaultColors(QtColorComboBox *colorComboBox);
	void UpdateSettings(const EchoChamberSettings& newSettings);
	void on_visualizationWindow_created(QWidget* object);
	QString GetTimelineColumn(void) const;
	// QString QueryForSQLEdit(QSqlQueryModelFetchAll* model, QString databaseName);
	QString QueryForTickEdit(AnimatedModel* model);
	static void SaveLayout(XMLSaver &xmlSaver);
	static void LoadAllWindows(XMLSaver &xmlSaver);

	static void SetCurrentText(QComboBox *comboBox, QString string);
	static bool GetImageFromModel(QModelIndex modelIndex, QImage *image);
	void RegenerateTimelineConstraints(void);

	QWorkspace *GetWorkspace(void) const {return workspace;}
	bool InitAMCWithDB(AnimatedModel *amc, QString query, QString databaseName);

	void SetWindowName(QWidget *window, const QString & windowType, QString windowNameSuffix);

	bool showTipsOnStartup;

public slots:
	void on_actionSave_Layout_activated();
	void on_actionLoad_Layout_activated();
	void on_actionSQL_Query_activated();
	void on_actionImage_Viewer_activated();
	void on_actionCascade_activated();
	void on_actionTile_activated();
	void on_actionQuery_Clipboard_activated();
	void on_actionTimeline_activated();
	void on_queryClipboard_itemDoubleClicked(QListWidgetItem *);
	void on_actionNew_Project_activated();
	void on_actionOpen_Project_activated();
	void on_actionSave_Project_activated();
	void on_actionSave_Project_As_activated();
		
	void on_actionFunction_Calls_activated();
	void on_actionPacket_Logs_activated();
	void on_actionNetwork_Statistics_activated();

	void on_actionEditSettings_activated();

	void onWindowClosed(QObject *o);
	void onWindowClosed(QWidget *o);
	void onWorkspaceWindowGotFocus(QWidget *);
	void on_timelineSlider_valueChanged(double timelineSliderValue);
	void on_timelineSlider_sliderMoved(double timelineSliderValue);

	void on_addLastQueryPushButton_released();
	void on_removeSelectedQueryPushButton_released();

	void on_stepBackwardsToolButton_released();
	void on_stepForwardsToolButton_released();
	void on_loopAnimationToolButton_released();
	void on_playAnimationToolButton_released();
	void on_stopAnimationToolButton_released();
	void animationTimerUpdate();
	void on_timestampColumnComboBox_currentIndexChanged(int idx);
	void on_visualizationWindow_destroyed(QObject* object);
	void on_actionLoad_Database_triggered();
	void on_actionUnload_Database_triggered();
	void on_sqlQueryTool_destroyed(QObject*);
	void on_timestampIntervalSpinBox_valueChanged(double d);
	void on_playbackFPSSpinBox_valueChanged(double d);
	void on_actionQuit_activated();

	void on_menu_Recent_Projects_triggered(QAction *action);
	void on_actionWhat_s_This_activated();
	void on_actionManual_activated();
	void on_actionWebsite_activated();
	void on_actionAbout_activated();

	const EchoChamberSettings& GetSettings(void) const {return settings;}

signals:
	void animationTick( bool isForward, int cur );


private:
	// Prevent crashes in destructor from destroyed() signal
	bool shuttingDown;

	QWorkspace *workspace;
	// Used to give each SQL Query window a unique name
	unsigned int windowSuffixInteger;
	// If true, call RebuildWindowListMenu next tick
	// Deferred so if this is set to true on destruction, the app doesn't crash
	bool needsUpdateWindowList;
	bool IsWidgetInWorkspace(QWidget *qw) const;
	bool IsWidgetInWorkspace(const QString &qs) const;
	static void TruncateAtCharacter(const QString &qs, QChar qc, QString &out);
	void RemoveFromWindowListMenu(QWidget *qw);
	void RebuildWindowListMenu(void);
	void SetupWindow(QWidget *window);
	void SpawnSQLQueryToolWithFirstPassingQuery(const QString &query);
	bool AddDatabase(QString dbName);
	void RemoveDatabase(QString dbName);
	void RegenerateTimestampColumn(int databaseIndex);
	void RegenerateTimestampColumns(void);
	void GetAnimatedModels(QVector<AnimatedModel*> &animatedModels);
	static void GetSaveableWindows(QVector<ISaveableWindow*> &saveableWindows);
	void SetAnimatedModelsToTimeline(void);
	void UpdateLabel(void);
	void UpdateLabelAndModels(void);
	QSqlQuery TestQuery(QString queryString, QString databaseName);
	void SaveProject(XMLSaver &xmlSaver);
	void LoadProject(XMLSaver &xmlSaver);
	QString LoadDatabase();
	void SetDefault(void);
	void OpenProject(QString fileName);

	QStringList databases;
	EchoChamberSettings settings;
	ScaleDrawNoScientific scaleDraw;
	QTimer animationTimer;
	QVector<SQLQueryTool*> sqlQueryTools;
	float timelineSliderIncrementPerAnimationTick;
	QString projectFilename;
	bool stateChangedSinceLastSave;

	void AddToRecentProjectList(QString name);
	void SaveRecentProjectList(void);
	void LoadRecentProjectList(void);
	void UpdateRecentProjectListWidget(void);
	QStringList recentProjects;
	QString mostRecentProjectPath;
	QString mostRecentDatabasePath;
	QString mostRecentLayoutPath;
	QTime whenAnimationStarted;

	void LoadAppSettings();
	void SaveAppSettings();
	
// 	bool InitAMCWithDB(AnimatedModelList *amc, QStringList queries, QStringList databaseNames);
};

#endif // ECHOCHAMBER_H
