#include <QtGui>

#include "ImageViewer.h"
#include "EchoChamber.h"
extern EchoChamber *echoChamber;

ImageViewer::ImageViewer(QWidget *parent) : QMainWindow(parent)
{
	imageLabel = new QLabel;
	imageLabel->setBackgroundRole(QPalette::Base);
	imageLabel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored);
	imageLabel->setScaledContents(true);

	scrollArea = new QScrollArea;
	scrollArea->setBackgroundRole(QPalette::Dark);
	scrollArea->setWidget(imageLabel);
	setCentralWidget(scrollArea);

	CreateActions();
	CreateMenus();

	setWindowIcon(QIcon("Resources/image_viewer.png"));

	setWindowTitle(tr("Image Viewer"));
	resize(500, 400);
}
bool ImageViewer::OpenFile(QString _fileName)
{
	QImage image;
	if (!_fileName.isEmpty())
	{
		if (image.load(_fileName)==false)
		{
			QMessageBox::information(echoChamber, tr("Image Viewer"),
				tr("Cannot load %1.").arg(fileName));
		}
		else
		{
			SetToImage(image);
			QMainWindow::statusBar()->showMessage(fileName);
			return true;
		}
	}
	return false;
}
void ImageViewer::OpenImageWithDialog(void)
{
	QImage image;
	QDir dir;
	if (fileName.isEmpty()==false)
	{
		dir = fileName;
		dir.cdUp();
	}
	QString _fileName = QFileDialog::getOpenFileName(echoChamber,
		tr("Open File"), dir.path(), "Image Files (*.png *.jpg *.bmp *.dds);;All Files (*.*)");
	if (!_fileName.isEmpty())
	{
		if (image.load(_fileName)==false)
		{
			QMessageBox::information(echoChamber, tr("Image Viewer"),
				tr("Cannot load %1.").arg(_fileName));
		}
		else
		{
			QMainWindow::statusBar()->showMessage(fileName);
			SetToImage(image);
		}
	}
}
bool ImageViewer::OpenImageWithDialog(QImage *image, QString &_fileName)
{
	_fileName = QFileDialog::getOpenFileName(echoChamber,
		tr("Open File"), QDir::currentPath(), "Image Files (*.png *.jpg *.bmp *.dds);;All Files (*.*)");
	if (!_fileName.isEmpty())
	{
		if (image->load(_fileName)==false)
		{
			QMessageBox::information(echoChamber, tr("Image Viewer"),
				tr("Cannot load %1.").arg(_fileName));
		}
		else
			return true;
	}
	return false;
}
void ImageViewer::SetToImage(QImage image, QString _fileName)
{
	fileName=_fileName;
	QMainWindow::statusBar()->showMessage(fileName);
	imageLabel->setPixmap(QPixmap::fromImage(image));
	scaleFactor = 1.0;

	saveAsAct->setEnabled(true);
	printAct->setEnabled(true);
	fitToWindowAct->setEnabled(true);
	UpdateActions();

	if (fitToWindowAct->isChecked())
		fitToWindow();
	else
		imageLabel->adjustSize();
}
void ImageViewer::open()
{
	OpenImageWithDialog();
}
void ImageViewer::saveAs()
{
	QDir dir;
	if (fileName.isEmpty()==false)
	{
		dir = fileName;
		dir.cdUp();
	}
	QString _fileName
		= QFileDialog::getSaveFileName(this, tr("Save Image"), dir.path(), "Image Files (*.png *.jpg *.bmp);;All Files (*.*)");
	if (_fileName.isEmpty())
		return;
	QFile file(_fileName);
	if (!file.open(QFile::WriteOnly)) {
		QString msg = tr("Failed to open %1\n%2")
			.arg(_fileName)
			.arg(file.errorString());
		QMessageBox::warning(this, tr("Error"), msg);
		return;
	}

	if (!imageLabel->pixmap()->save(&file))
	{
		file.close();
		file.remove(_fileName);
		QMessageBox::warning(this, tr("Error"), "Unknown image format extension. Use .png, .jpg, or .bmp");
	}
	else
	{
		fileName=_fileName;
		QMainWindow::statusBar()->showMessage(fileName);
	}
}
void ImageViewer::print()
{
	Q_ASSERT(imageLabel->pixmap());
	QPrintDialog dialog(&printer, this);
	if (dialog.exec()) {
		QPainter painter(&printer);
		QRect rect = painter.viewport();
		QSize size = imageLabel->pixmap()->size();
		size.scale(rect.size(), Qt::KeepAspectRatio);
		painter.setViewport(rect.x(), rect.y(), size.width(), size.height());
		painter.setWindow(imageLabel->pixmap()->rect());
		painter.drawPixmap(0, 0, *imageLabel->pixmap());
	}
}

void ImageViewer::zoomIn()
{
	ScaleImage(2.0);
}

void ImageViewer::zoomOut()
{
	ScaleImage(0.5);
}

void ImageViewer::normalSize()
{
	imageLabel->adjustSize();
	scaleFactor = 1.0;
}

void ImageViewer::fitToWindow()
{
	bool fitToWindow = fitToWindowAct->isChecked();
	scrollArea->setWidgetResizable(fitToWindow);
	if (!fitToWindow) {
		normalSize();
	}
	UpdateActions();
}
void ImageViewer::CreateActions()
{
	openAct = new QAction(tr("&Open..."), this);
	openAct->setShortcut(tr("Ctrl+O"));
	connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

	saveAsAct = new QAction(tr("&Save..."), this);
	saveAsAct->setShortcut(tr("Ctrl+S"));
	saveAsAct->setEnabled(false);
	connect(saveAsAct, SIGNAL(triggered()), this, SLOT(saveAs()));

	printAct = new QAction(tr("&Print..."), this);
	printAct->setShortcut(tr("Ctrl+P"));
	printAct->setEnabled(false);
	connect(printAct, SIGNAL(triggered()), this, SLOT(print()));

	exitAct = new QAction(tr("E&xit"), this);
	exitAct->setShortcut(tr("Ctrl+Q"));
	connect(exitAct, SIGNAL(triggered()), this, SLOT(close()));

	zoomInAct = new QAction(tr("Zoom &In (25%)"), this);
	zoomInAct->setShortcut(tr("Ctrl++"));
	zoomInAct->setEnabled(false);
	connect(zoomInAct, SIGNAL(triggered()), this, SLOT(zoomIn()));

	zoomOutAct = new QAction(tr("Zoom &Out (25%)"), this);
	zoomOutAct->setShortcut(tr("Ctrl+-"));
	zoomOutAct->setEnabled(false);
	connect(zoomOutAct, SIGNAL(triggered()), this, SLOT(zoomOut()));

	normalSizeAct = new QAction(tr("&Normal Size"), this);
	normalSizeAct->setShortcut(tr("Ctrl+S"));
	normalSizeAct->setEnabled(false);
	connect(normalSizeAct, SIGNAL(triggered()), this, SLOT(normalSize()));

	fitToWindowAct = new QAction(tr("&Fit to Window"), this);
	fitToWindowAct->setEnabled(false);
	fitToWindowAct->setCheckable(true);
	fitToWindowAct->setShortcut(tr("Ctrl+F"));
	connect(fitToWindowAct, SIGNAL(triggered()), this, SLOT(fitToWindow()));

}

void ImageViewer::CreateMenus()
{
	fileMenu = new QMenu(tr("&File"), this);
	fileMenu->addAction(openAct);
	fileMenu->addAction(saveAsAct);
	fileMenu->addAction(printAct);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAct);

	viewMenu = new QMenu(tr("&View"), this);
	viewMenu->addAction(zoomInAct);
	viewMenu->addAction(zoomOutAct);
	viewMenu->addAction(normalSizeAct);
	viewMenu->addSeparator();
	viewMenu->addAction(fitToWindowAct);

	menuBar()->addMenu(fileMenu);
	menuBar()->addMenu(viewMenu);
}

void ImageViewer::UpdateActions()
{
	zoomInAct->setEnabled(!fitToWindowAct->isChecked());
	zoomOutAct->setEnabled(!fitToWindowAct->isChecked());
	normalSizeAct->setEnabled(!fitToWindowAct->isChecked());
}

void ImageViewer::ScaleImage(double amount)
{
	Q_ASSERT(imageLabel->pixmap());
	scaleFactor *= amount;
	imageLabel->resize(scaleFactor * imageLabel->pixmap()->size());

	AdjustScrollBar(scrollArea->horizontalScrollBar(), amount);
	AdjustScrollBar(scrollArea->verticalScrollBar(), amount);

	zoomInAct->setEnabled(scaleFactor < 3.0);
	zoomOutAct->setEnabled(scaleFactor > 0.333);
}

void ImageViewer::AdjustScrollBar(QScrollBar *scrollBar, double factor)
{
	scrollBar->setValue(int(factor * scrollBar->value()
		+ ((factor - 1) * scrollBar->pageStep()/2)));
}
void ImageViewer::Save(XMLSaver &xmlSaver)
{
	// this is always first
	SaveType(xmlSaver);
	SaveWindowTitle(xmlSaver,this);
	SaveVersion(xmlSaver);



	QByteArray geo_data = saveGeometry();
	QByteArray layout_data = saveState();
	XML_SAVER_SAVE_ELEMENT(xmlSaver, geo_data);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, layout_data);

	bool fitToWindowBool = fitToWindowAct->isChecked();
	XML_SAVER_SAVE_ELEMENT(xmlSaver, fitToWindowBool);

	xmlSaver.SaveElement("fileName", fileName);
}
void ImageViewer::Load(XMLSaver &xmlSaver)
{
	int version;
	LoadVersion(xmlSaver, version);

	QByteArray geo_data;
	QByteArray layout_data;
	XML_SAVER_LOAD_ELEMENT(xmlSaver, geo_data, false);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, layout_data, false);
	restoreGeometry(geo_data);
	restoreState(layout_data);

	bool fitToWindowBool;
	XML_SAVER_LOAD_ELEMENT(xmlSaver, fitToWindowBool, false);
	if (fitToWindowBool!=fitToWindowAct->isChecked())
	{
		fitToWindowAct->setChecked(fitToWindowBool);
		if (fitToWindowAct->isEnabled())
		{
			if (fitToWindowAct->isEnabled())
				fitToWindow();
		}
	}

	QString testFilename;
	xmlSaver.LoadElement("fileName", testFilename, "");
	OpenFile(testFilename);
}
