#include "PostfixTableRowQuery.h"
#include "EchoChamber.h"
#include <QStringList>
#include <assert.h>

extern EchoChamber *echoChamber;

PostfixTableRowQuery::PostfixTableRowQuery()
{
}
PostfixTableRowQuery::~PostfixTableRowQuery()
{
}
bool PostfixTableRowQuery::SetQuery(const QString &query)
{
	QList<QueryToken> testQuery;
	if (EvaluateQuery(query, testQuery))
	{
		currentQuery=testQuery;
		queryString=query;
		return true;
	}
	return false;
}
bool PostfixTableRowQuery::EvaluateQuery(const QString &query, QList<QueryToken> &tokenList)
{
	tokenList.clear();
	if (query.isEmpty())
		return true;

	QStringList tokenStrings = query.split(POSTFIX_TABLE_ROW_TOKEN_SEPARATOR, QString::SkipEmptyParts,
		Qt::CaseInsensitive);
	if (tokenStrings.size()==0)
		return false;
	for (int i=0; i < tokenStrings.size(); i++)
	{
		if (tokenStrings[i].toUpper()==POSTFIX_TABLE_ROW_QUERY_AND)
		{
			// If prior row is none, next row must be unary operator
			if (tokenList.size()>0 && tokenList[tokenList.size()-1].operation==QueryToken::NONE)
				return false;
			tokenList.push_back(QueryToken(QString(),QueryToken::AND));
		}
		else if (tokenStrings[i].toUpper()==POSTFIX_TABLE_ROW_QUERY_OR)
		{
			// If prior row is none, next row must be unary operator
			if (tokenList.size()>0 && tokenList[tokenList.size()-1].operation==QueryToken::NONE)
				return false;
			tokenList.push_back(QueryToken(QString(),QueryToken::OR));
		}
		else if (tokenStrings[i].toUpper()==POSTFIX_TABLE_ROW_QUERY_CHANGED)
		{
			// Prior row must be a column name for unary operator
			if (tokenList.size()==0 || tokenList[tokenList.size()-1].operation!=QueryToken::NONE)
				return false;
			QString priorStr = tokenList[tokenList.size()-1].columnName;
			tokenList.pop_back();
			tokenList.push_back(QueryToken(priorStr,QueryToken::COLUMN_CHANGED));
		}
		else if (tokenStrings[i].toUpper()==POSTFIX_TABLE_ROW_QUERY_SAME)
		{
			// Prior row must be a column name for unary operator
			if (tokenList.size()==0 || tokenList[tokenList.size()-1].operation!=QueryToken::NONE)
				return false;
			QString priorStr = tokenList[tokenList.size()-1].columnName;
			tokenList.pop_back();
			tokenList.push_back(QueryToken(priorStr,QueryToken::COLUMN_SAME));
		}
		else
		{
			tokenList.push_back(QueryToken(tokenStrings[i],QueryToken::NONE));
			// can't end with just a column name
			if (i==tokenStrings.size()-1)
				return false;
		}
	}

	int argCount=0;
	for (int i=0; i < tokenList.size(); i++)
	{
		if (tokenList[i].operation==QueryToken::AND ||
			tokenList[i].operation==QueryToken::OR)
		{
			// Consumes 2, adds 1
			if (argCount<2)
				return false;
			argCount--;
		}
		else
			argCount++;
	}
	if (argCount!=1)
		return false;

	return true;
}
bool PostfixTableRowQuery::Preparse(const QSqlQueryModelFetchAll *model)
{
	// Set unused rows to QTRUE or QFALSE, and do index lookups
	for (int i=0; i < currentQuery.size(); i++)
	{
		if ((currentQuery[i].operation==QueryToken::COLUMN_CHANGED || currentQuery[i].operation==QueryToken::COLUMN_SAME))
		{
			currentQuery[i].columnIndex=EchoChamber::GetColumnIndex(model, currentQuery[i].columnName);
			if (currentQuery[i].columnIndex==-1)
			{
				return false;
			}
		}
	}
	return true;
}
int PostfixTableRowQuery::GetNextRowStepIndexInclusive(QSqlQueryModelFetchAll *model, int lastUsedRow, bool forward)
{
	int increment;
	int stop;
	if (forward)
	{
		increment=1;
		stop=model->realRowCount();
	}
	else
	{
		increment=-1;
		stop=-1;
	}
	if (lastUsedRow==stop)
		return lastUsedRow;

	int rowIndex;
	for (rowIndex=lastUsedRow+increment; rowIndex != stop; rowIndex+=increment)
	{
		if (ExpressionPassed(model, rowIndex, lastUsedRow))
			break;
	}

	return rowIndex-increment;
}
void PostfixTableRowQuery::EvalColumn(const QSqlQueryModelFetchAll *model, int curRow, int otherRow, int curIndex, QList<PostfixTableRowQuery::QueryToken> &workingQuery)
{
	if (workingQuery[curIndex].operation==QueryToken::COLUMN_CHANGED)
	{
		if (ColumnChanged(model,curRow,otherRow, workingQuery[curIndex].columnIndex))
			workingQuery[curIndex].operation=QueryToken::QTRUE;
		else
			workingQuery[curIndex].operation=QueryToken::QFALSE;
	}
	else if (workingQuery[curIndex].operation==QueryToken::COLUMN_SAME)
	{
		if (ColumnChanged(model,curRow,otherRow, workingQuery[curIndex].columnIndex))
			workingQuery[curIndex].operation=QueryToken::QFALSE;
		else
			workingQuery[curIndex].operation=QueryToken::QTRUE;
	}
}
bool PostfixTableRowQuery::ExpressionPassed(const QSqlQueryModelFetchAll *model, int curRow, int otherRow)
{
	QList<QueryToken> workingQuery=currentQuery;
	int curIndex=0;
	while (workingQuery.size()>1)
	{
		// Look for operators AND and OR
		if (workingQuery[curIndex].operation==QueryToken::AND)
		{
			assert(curIndex>=2);

			if (workingQuery[curIndex-1].operation!=QueryToken::QFALSE &&
				workingQuery[curIndex-2].operation!=QueryToken::QFALSE)
			{
				// Eval to QTRUE or QFALSE for prior two
				EvalColumn(model, curRow, otherRow, curIndex-1, workingQuery);
				if (workingQuery[curIndex-1].operation==QueryToken::QTRUE)
					EvalColumn(model, curRow, otherRow, curIndex-2, workingQuery);
			}

			if (workingQuery[curIndex-1].operation==QueryToken::QTRUE &&
				workingQuery[curIndex-2].operation==QueryToken::QTRUE)
			{
				// Pop 3
				// Push QTRUE
				workingQuery.removeAt(curIndex);
				workingQuery.removeAt(curIndex-1);
				workingQuery[curIndex-2].operation=QueryToken::QTRUE;
				workingQuery[curIndex-2].columnName.clear();
				curIndex=curIndex-2;
			}
			else
			{
				// Pop 3
				// Push QFALSE
				workingQuery.removeAt(curIndex);
				workingQuery.removeAt(curIndex-1);
				workingQuery[curIndex-2].operation=QueryToken::QFALSE;
				workingQuery[curIndex-2].columnName.clear();
				curIndex=curIndex-2;
			}
		}
		else if (workingQuery[curIndex].operation==QueryToken::OR)
		{
			assert(curIndex>=2);

			if (workingQuery[curIndex-1].operation!=QueryToken::QTRUE &&
				workingQuery[curIndex-2].operation!=QueryToken::QTRUE)
			{
				EvalColumn(model, curRow, otherRow, curIndex-1, workingQuery);
				if (workingQuery[curIndex-1].operation!=QueryToken::QTRUE)
					EvalColumn(model, curRow, otherRow, curIndex-2, workingQuery);
			}

			if (workingQuery[curIndex-1].operation==QueryToken::QTRUE ||
				workingQuery[curIndex-2].operation==QueryToken::QTRUE)
			{
				// Pop 3
				// Push QTRUE
				workingQuery.removeAt(curIndex);
				workingQuery.removeAt(curIndex-1);
				workingQuery[curIndex-2].operation=QueryToken::QTRUE;
				workingQuery[curIndex-2].columnName.clear();
				curIndex=curIndex-2;
			}
			else
			{
				// Pop 3
				// Push QFALSE
				workingQuery.removeAt(curIndex);
				workingQuery.removeAt(curIndex-1);
				workingQuery[curIndex-2].operation=QueryToken::QFALSE;
				workingQuery[curIndex-2].columnName.clear();
				curIndex=curIndex-2;
			}
		}
		else
		{
			// Shouldn't go off the end due to prior validity checks
			curIndex++;
			assert(curIndex!=workingQuery.size());
		}
	}

	if (workingQuery[0].operation==QueryToken::QTRUE)
		return true;
	return false;
}
bool PostfixTableRowQuery::ColumnChanged(const QSqlQueryModelFetchAll *model, int curRow, int otherRow, int column) const
{
	QVariant curRowVariant = model->data(model->index(curRow, column));
	QVariant otherRowVariant = model->data(model->index(otherRow, column));
	return curRowVariant!=otherRowVariant;
}
void PostfixTableRowQuery::Save(XMLSaver &xmlSaver)
{
	XML_SAVER_SAVE_ELEMENT(xmlSaver, queryString);
}
void PostfixTableRowQuery::Load(XMLSaver &xmlSaver)
{
	QString q;
	xmlSaver.LoadElement("queryString", q, "");
	SetQuery(q);
}
