/* 
 * File:   QDataTypes3D.h
 *
 * Created on September 13, 2009, 3:47 PM
 */

#ifndef _QDATATYPES3D_H
#define	_QDATATYPES3D_H

#include <float.h>
#include <math.h>
#include <assert.h>
#include "QtCore/QtGlobal"

class QPoint3F
{
public:
    float x;
    float y;
    float z;

    explicit QPoint3F() : x(0.0f), y(0.0f), z(0.0f) { }

    QPoint3F(float inx, float iny, float inz) : x(inx), y(iny), z(inz) { }

    QPoint3F(const QPoint3F & src) : x(src.x), y(src.y), z(src.z) { }

    QPoint3F(QPoint3F & src) : x(src.x), y(src.y), z(src.z) { }

    QPoint3F & operator =(const QPoint3F& input)
    {
        x = input.x;
        y = input.y;
        z = input.z;
        return *this;
    }

    inline void set(float inx, float iny, float inz)
    {
        x = inx;
        y = iny;
        z = inz;
    }

    inline void setX(float inx)
    {
        x = inx;
    }

    inline void setY(float iny)
    {
        y = iny;
    }

    inline void setZ(float inz)
    {
        z = inz;
    }

    inline float getX() const
    {
        return x;
    }

    inline float getY() const
    {
        return y;
    }

    inline float getZ() const
    {
        return z;
    }

    inline const QPoint3F operator -(const QPoint3F & other) const
    {
        return QPoint3F(x - other.x, y - other.y, z - other.z);
    }

    inline const QPoint3F operator +(const QPoint3F & other) const
    {
        return QPoint3F(x + other.x, y + other.y, z + other.z);
    }

    inline const QPoint3F operator /(float f) const
    {
        return QPoint3F(x / f, y / f, z / f);
    }

    inline float Dot(const QPoint3F & other) const
    {
        return x * other.x + y * other.y + z * other.z;
    }

    inline float Length(void) const
    {
        return sqrtf(x * x + y * y + z * z);
    }

    inline float Normalize()
    {
        float length = Length( );
        if (length != 0.0f)
        {
            float tmpLen = 1.0f / length;
            x *= tmpLen;
            y *= tmpLen;
            z *= tmpLen;
        }
        else
        {
            x = 0.0f;
            y = 0.0f;
            z = 0.0f;
        }
        return length;
    }

    inline float LengthSq(void) const
    {
        return x * x + y * y + z*z;
    }

    static QPoint3F center(const QPoint3F & min, const QPoint3F & max)
    {
        return QPoint3F((min.x + max.x) * 0.5f, (min.y + max.y) * 0.5f, (min.z + max.z) * 0.5f);
    }
};

class QAxisAlignedBoundingBox
{
private:
    // Don't use variable names just 'min' and 'max' as these are macros
    QPoint3F minPoint;
    QPoint3F maxPoint;
    bool valid;

protected:

public:

    explicit QAxisAlignedBoundingBox() : minPoint(), maxPoint(), valid(true) { }

    explicit QAxisAlignedBoundingBox(const QAxisAlignedBoundingBox & copy) : minPoint(copy.minPoint), maxPoint(copy.maxPoint), valid(copy.valid) { }

    QAxisAlignedBoundingBox(const QPoint3F & inmin, const QPoint3F & inmax) : minPoint(inmin), maxPoint(inmax), valid(true) { }

    QAxisAlignedBoundingBox & operator =(const QAxisAlignedBoundingBox& input)
    {
        minPoint = input.minPoint;
        maxPoint = input.maxPoint;
        valid = input.valid;
        return *this;
    }

    inline void reset()
    {
        minPoint.set(FLT_MAX, FLT_MAX, FLT_MAX);
        maxPoint.set(-FLT_MAX, -FLT_MAX, -FLT_MAX);
        valid = false;
    }

    inline bool isValid() const { return valid; }

    inline const QPoint3F getSquareMin() const
    {
        float maxExtentX = maxPoint.getX() - minPoint.getX();
        float maxExtentY = maxPoint.getY() - minPoint.getY();
        float maxExtentZ = maxPoint.getZ() - minPoint.getZ();
        float maxExtent = maxExtentX;
        if (maxExtentY > maxExtent)
            maxExtent = maxExtentY;
        if (maxExtentZ > maxExtent)
            maxExtent = maxExtentZ;
        return QPoint3F(minPoint.getX() + maxExtentX / 2 - maxExtent / 2, minPoint.getY() + maxExtentY / 2 - maxExtent / 2, minPoint.getZ() + maxExtentZ / 2 - maxExtent / 2);
    }

    inline const QPoint3F getSquareMax() const
    {
        float maxExtentX = maxPoint.getX() - minPoint.getX();
        float maxExtentY = maxPoint.getY() - minPoint.getY();
        float maxExtentZ = maxPoint.getZ() - minPoint.getZ();
        float maxExtent = maxExtentX;
        if (maxExtentY > maxExtent)
            maxExtent = maxExtentY;
        if (maxExtentZ > maxExtent)
            maxExtent = maxExtentZ;
        return QPoint3F(minPoint.getX() + maxExtentX / 2 + maxExtent / 2, minPoint.getY() + maxExtentY / 2 + maxExtent / 2, minPoint.getZ() + maxExtentZ / 2 + maxExtent / 2);
    }

    inline const QPoint3F & getMin() const
    {
        return minPoint;
    }

    inline const QPoint3F & getMax() const
    {
        return maxPoint;
    }

    inline const QPoint3F getCenter() const
    {
        return QPoint3F::center(minPoint, maxPoint);
    }

    inline float getSquareWidth() const
    {
        float maxExtentX = maxPoint.getX() - minPoint.getX();
        float maxExtentY = maxPoint.getY() - minPoint.getY();
        float maxExtentZ = maxPoint.getZ() - minPoint.getZ();
        float maxExtent = maxExtentX;
        if (maxExtentY > maxExtent)
            maxExtent = maxExtentY;
        if (maxExtentZ > maxExtent)
            maxExtent = maxExtentZ;
        return maxExtent;
    }

    inline float getSquareHeight() const
    {
        return getSquareWidth();
    }

    inline float getSquareDepth() const
    {
        return getSquareWidth();
    }

    inline float getWidth() const
    {
        return maxPoint.getX() - minPoint.getX();
    }

    inline float getHeight() const
    {
        return maxPoint.getY() - minPoint.getY();
    }

    inline float getDepth() const
    {
        return maxPoint.getZ() - minPoint.getZ();
    }

    inline QPoint3F getSquareDelta() const
    {
        float dist = getSquareWidth();
        return QPoint3F(dist, dist, dist);
    }

    inline QPoint3F getDelta() const
    {
        return getMax() - getMin();
    }

    /** Adds a point to this bounding box, possibly extending the min and max limits to contain the point.
     * param x X coord of the point to be added.
     * param y Y coord of the point to be added.
     * param z Z coord of the point to be added.
     */
    inline void add(const float x, const float y, const float z)
    {
        valid = true;
        if (x < minPoint.getX()) minPoint.setX(x);
        else if (x > maxPoint.getX()) maxPoint.setX(x);

        if (y < minPoint.getY()) minPoint.setY(y);
        else if (y > maxPoint.getY()) maxPoint.setY(y);

        if (z < minPoint.getZ()) minPoint.setZ(z);
        else if (z > maxPoint.getZ()) maxPoint.setZ(z);
    }

    /** Adds a point to this bounding box, possibly extending the min and max limits to contain the point.
     * param point The other bounding box to add to this one.
     * see QPoint3F
     */
    inline void add(const QPoint3F & point)
    {
        add(point.getX(), point.getY(), point.getZ());
    }

    /** Adds another bounding box to this one, possibly extending the min and max points.
     * param otherBounds The other bounding box to add to this one.
     */
    inline void add(const QAxisAlignedBoundingBox & otherBounds)
    {
        if (otherBounds.isValid())
        {
            add(otherBounds.getMin());
            add(otherBounds.getMax());
        }
    }
};

class QPlane
{
private:
    QPoint3F planePoint;
    QPoint3F planeNormal;

public:

    explicit QPlane(const QPoint3F &point, const QPoint3F &normal) : planePoint(point), planeNormal(normal) { }

    float PointToPlaneDist(const QPoint3F &point)
    {
        return sqrt(PointToPlaneDistSq(point));
    }

    float PointToPlaneDistSq(const QPoint3F &point)
    {
        return (point - planePoint).Dot(planeNormal);
    }

    QPlane(const QPlane & copy) : planePoint(copy.planePoint), planeNormal(copy.planeNormal) { }

    QPlane & operator =(const QPlane& input)
    {
        planePoint = input.planePoint;
        planeNormal = input.planeNormal;
        return *this;
    }
};


class QMatrix4x4
{
private:
public:
    float m[16];

    QMatrix4x4()
    {
    }
    ~QMatrix4x4() {}
    void identity()
    {
        m[ 0 ] = m[ 5 ] = m[ 10 ] = m[ 15 ] = 1.0f;
        m[ 1 ] = m[ 2 ] = m[ 3 ] = m[ 4 ] = 0.0f;
        m[ 6 ] = m[ 7 ] = m[ 8 ] = m[ 9 ] = 0.0f;
        m[ 11 ] = m[ 12 ] = m[ 13 ] = m[ 14 ] = 0.0f;
    }

    QPoint3F rotate(const QPoint3F & point) const
    {
        return QPoint3F( (m[ 0 ] * point.getX()) + (m[4] * point.getY()) + (m[8] * point.getZ()), (m[ 1 ] * point.getX()) + (m[5] * point.getY()) + (m[9] * point.getZ()), (m[ 2 ] * point.getX()) + (m[6] * point.getY()) + (m[10] * point.getZ()));
    }
    
    QPoint3F transform(const QPoint3F & point) const
    {
        return QPoint3F( (m[ 0 ] * point.getX()) + (m[4] * point.getY()) + (m[8] * point.getZ()) + m[12], (m[ 1 ] * point.getX()) + (m[5] * point.getY()) + (m[9] * point.getZ()) + m[13], (m[ 2 ] * point.getX()) + (m[6] * point.getY()) + (m[10] * point.getZ()) + m[14]);
    }

    float determinant() const
    {
	return m[0] * (m[5] * m[10] - m[6] * m[9]) + m[1] * (m[6] * m[8] - m[4] * m[10]) + m[2] * (m[4] * m[9] - m[5] * m[8]);
    }

    QMatrix4x4 inverse() const
    {
        QMatrix4x4 outMatrix;

	float	deter = determinant();

	assert( fabs( deter ) > 0.000001f );

	float	determinant_inv = 1.0f / deter;

	// Compute the 3x3 portion of the inverse, by dividing the adjoint by the determinant

	outMatrix.m[0] = (m[5] * m[10] - m[6] * m[9]) * determinant_inv;
	outMatrix.m[1] = (m[2] * m[9] - m[1] * m[10]) * determinant_inv;
	outMatrix.m[2] = (m[1] * m[6] - m[2] * m[5]) * determinant_inv;

	outMatrix.m[4] = (m[6] * m[8] - m[4] * m[10]) * determinant_inv;
	outMatrix.m[5] = (m[0] * m[10] - m[2] * m[8]) * determinant_inv;
	outMatrix.m[6] = (m[2] * m[4] - m[0] * m[6]) * determinant_inv;

	outMatrix.m[8] = (m[4] * m[9] - m[5] * m[8]) * determinant_inv;
	outMatrix.m[9] = (m[1] * m[8] - m[0] * m[9]) * determinant_inv;
	outMatrix.m[10] = (m[0] * m[5] - m[1] * m[4]) * determinant_inv;

	// Compute the translation portion of the inverse

	outMatrix.m[12] = -(m[12] * outMatrix.m[0] + m[13] * outMatrix.m[4] + m[14] * outMatrix.m[8]);
	outMatrix.m[13] = -(m[12] * outMatrix.m[1] + m[13] * outMatrix.m[5] + m[14] * outMatrix.m[9]);
	outMatrix.m[14] = -(m[12] * outMatrix.m[2] + m[13] * outMatrix.m[6] + m[14] * outMatrix.m[10]);

        return outMatrix;
    }
};

#endif	/* _QDATATYPES3D_H */

