#include "QwtWindowTab.h"
#include "EchoChamber.h"
#include "QwtWindow.h"
#include <qwt_plot.h>
#include <assert.h>
#include "histogram_item.h"
#include <qwt_plot_curve.h>
#include <QPen>
#include <qwt_symbol.h>
#include <float.h>
#include <qwt_interval_data.h>
#include <assert.h>
#include <qwt_painter.h>
#include <QPainter>
#include <QBitArray>

extern EchoChamber *echoChamber;

static const int LINE_INDEX=0;
static const int HISTOGRAM_INDEX=1;
static const int SYMBOL_INDEX=2;

enum SymbolSizeAlgorithm
{
	SSA_BY_RADIUS,
	SSA_BY_DIAMETER,
	SSA_BY_SQUARE_AREA,
	SSA_BY_CUBE_VOLUME,
	SSA_BY_CIRCLE_AREA,
	SSA_BY_SPHERE_VOLUME,
	SSA_COUNT,
};
static const char* SYMBOL_ALGORITHMS[SSA_COUNT]={
	"Radius",
	"Diameter",
	"Square Area",
	"Cube Volume",
	"Circle Area",
	"Sphere Volume",
};

// Input is final output (volume, length, etc)
// Output is input to get that volume
double ApplyAlgorithm(SymbolSizeAlgorithm algorithm, double d)
{
	switch (algorithm)
	{
	case SSA_BY_RADIUS:
		// circle radius is d
		// r = d/2
		// Find r
		// r = 2 * d
		d*=2;
		break;
	case SSA_BY_DIAMETER:
		break;
	case SSA_BY_SQUARE_AREA:
		d=pow(d,1.0/2.0);
		break;
	case SSA_BY_CUBE_VOLUME:
		d=pow(d,1.0/3.0);
		break;
	case SSA_BY_CIRCLE_AREA:
		// circle volume is d
		// d = PI * r^2
		// find r
		// r = sqrt(d / PI)
		d = sqrt(d/M_PI);
		break;
	case SSA_BY_SPHERE_VOLUME:
		// Sphere volume is d
		// d = 4.0 / 3.0 * M_PI * (r/2) * (r/2) * (r/2);
		// find r
		// (r/2)^3 = 3d / (4*PI)
		// r/2 = (3d / (4*PI))^(1/3)
		// r = 2 * (3d / (4*PI))^(1/3)
		d = 2.0 * pow(3.0 * d / (4.0 * M_PI),1.0/3.0);
		break;
	}
	return d;
}

QwtWindowTab::QwtWindowTab(QWidget* parent, Qt::WindowFlags f, QwtWindow *_qwtWindow, QString query, QString database)
	: QWidget(parent,f), qwtPlot(_qwtWindow->qwtPlot), qwtWindow(_qwtWindow)
{
	setupUi(this);

	plotCurve = new QwtSymbolCurve;
	histogramItem = new HistogramItem;
	symbol = new QwtSymbol;
	pen = new QPen;

	inUpdateDataSource=false;

	echoChamber->AddDefaultColors(colorComboBox);

	penStyleComboBox->addItem("NoPen");
	penStyleComboBox->addItem("SolidLine");
	penStyleComboBox->addItem("DashLine");
	penStyleComboBox->addItem("DotLine");
	penStyleComboBox->addItem("DashDotLine");
	penStyleComboBox->addItem("DashDotDotLine");
	penStyleComboBox->setCurrentIndex(1);

	curveStyleComboBox->addItem("NoCurve");
	curveStyleComboBox->addItem("Lines");
	curveStyleComboBox->addItem("Sticks");
	curveStyleComboBox->addItem("Steps");
	curveStyleComboBox->addItem("Dots");
	curveStyleComboBox->setCurrentIndex(1);

	symbolStyleComboBox->addItem("No Symbol");
	symbolStyleComboBox->addItem("Ellipse");
	symbolStyleComboBox->addItem("Rect");
	symbolStyleComboBox->addItem("Diamond");
	symbolStyleComboBox->addItem("Triangle");
	symbolStyleComboBox->addItem("DTriangle");
	symbolStyleComboBox->addItem("UTriangle");
	symbolStyleComboBox->addItem("LTriangle");
	symbolStyleComboBox->addItem("RTriangle");
	symbolStyleComboBox->addItem("Cross");
	symbolStyleComboBox->addItem("XCross");
	symbolStyleComboBox->addItem("HLine");
	symbolStyleComboBox->addItem("VLine");
	symbolStyleComboBox->addItem("Star1");
	symbolStyleComboBox->addItem("Star2");
	symbolStyleComboBox->addItem("Hexagon");

	symbolStyleComboBox_2->addItem("Ellipse");
	symbolStyleComboBox_2->addItem("Rect");
	symbolStyleComboBox_2->addItem("Diamond");
	symbolStyleComboBox_2->addItem("Triangle");
	symbolStyleComboBox_2->addItem("DTriangle");
	symbolStyleComboBox_2->addItem("UTriangle");
	symbolStyleComboBox_2->addItem("LTriangle");
	symbolStyleComboBox_2->addItem("RTriangle");
	symbolStyleComboBox_2->addItem("Cross");
	symbolStyleComboBox_2->addItem("XCross");
	symbolStyleComboBox_2->addItem("HLine");
	symbolStyleComboBox_2->addItem("VLine");
	symbolStyleComboBox_2->addItem("Star1");
	symbolStyleComboBox_2->addItem("Star2");
	symbolStyleComboBox_2->addItem("Hexagon");

	for (int i=0; i < SSA_COUNT; i++)
		symbolSizeAlgorithmComboBox->addItem(SYMBOL_ALGORITHMS[i]);

	histogramAlignmentComboBox->addItem("y=f(x)");
	histogramAlignmentComboBox->addItem("x=f(y)");

	lineAlignmentComboBox->addItem("y=f(x)");
	lineAlignmentComboBox->addItem("x=f(y)");

	BuildPlotCurveFromSettings();
	histogramItem->setColor(colorComboBox->currentColor());

	if (query.size()>0)
	{
		PostDBInit(query, database);
	}
}
void QwtWindowTab::PostDBInit(QString query, QString database)
{
	echoChamber->InitAMCWithDB(this, query, database );
	echoChamber->SetupSymbolMagnitudeComboBox(GetModel(), symbolMagnitudeComboBox);
	int defaults = echoChamber->SetupXYZTimeComboBoxes(GetModel(), xColumnComboBox, yColumnComboBox, 0, 0);
	if (defaults&1==0)
	{
		echoChamber->SetupXYZTimeComboBoxes(GetModel(), 0, 0, 0, xColumnComboBox);
	}
	UpdateDataSource(true);

	if (tabWidget->currentIndex()==LINE_INDEX || tabWidget->currentIndex()==SYMBOL_INDEX)
	{
		plotCurve->detach();
		plotCurve->show();
		plotCurve->attach(qwtPlot);
		histogramItem->hide();
		histogramItem->detach();
	}
	else if (tabWidget->currentIndex()==HISTOGRAM_INDEX)
	{
		plotCurve->hide();
		plotCurve->detach();
		histogramItem->show();
		histogramItem->attach(qwtPlot);
	}
}
void QwtWindowTab::RebuildAll(void)
{
//	DetachCurveList();
	echoChamber->SetupSymbolMagnitudeComboBox(GetModel(), symbolMagnitudeComboBox);
	echoChamber->SetupXYZTimeComboBoxes(GetModel(), xColumnComboBox, yColumnComboBox, 0, 0);
	UpdateBasedOnTabWidget();
}
QwtWindowTab::~QwtWindowTab()
{
	DeleteAll();
}
void QwtWindowTab::DeleteAll()
{
//	int i;
//	for (i=0; i < plotList.size(); i++)
//	{
//		plotList[i]->detach();
//		delete plotList[i];
//	}
	delete plotCurve;
	plotCurve=0;
	delete histogramItem;
	histogramItem=0;
	delete symbol;
	symbol=0;
	delete pen;
	pen=0;
}
void QwtWindowTab::on_xColumnComboBox_activated(QString string)
{
	UpdateDataSource(true);
	qwtWindow->Replot();
}
void QwtWindowTab::on_yColumnComboBox_activated(QString string)
{
	UpdateDataSource(true);
	qwtWindow->Replot();
}
void QwtWindowTab::BuildPlotCurveFromSettings(void)
{
	pen->setWidth(linePenWidthSpinBox->value());
	if (tabWidget->currentIndex()==LINE_INDEX)
	{
		pen->setStyle((Qt::PenStyle) penStyleComboBox->currentIndex());
		plotCurve->setStyle((QwtPlotCurve::CurveStyle)curveStyleComboBox->currentIndex());
	}
	else if (tabWidget->currentIndex()==SYMBOL_INDEX)
	{
		pen->setStyle(Qt::NoPen);
		plotCurve->setStyle(QwtPlotCurve::NoCurve);
	}
	pen->setColor(colorComboBox->currentColor());
	plotCurve->setPen(*pen);
	symbol->setPen(colorComboBox->currentColor());
	if (tabWidget->currentIndex()==LINE_INDEX)
		symbol->setStyle((QwtSymbol::Style)(symbolStyleComboBox->currentIndex()-1));
	else if (tabWidget->currentIndex()==SYMBOL_INDEX)
		symbol->setStyle((QwtSymbol::Style)(symbolStyleComboBox_2->currentIndex()));
	symbol->setSize(symbolSizeDoubleSpinBox->value());
	plotCurve->setSymbol(*symbol);

	if (curveStyleComboBox->currentIndex()==0)
	{
		penStyleComboBox->setEnabled(false);
		linePenWidthSpinBox->setEnabled(false);
	}
	else
	{
		penStyleComboBox->setEnabled(true);
		linePenWidthSpinBox->setEnabled(true);
	}

	if (symbolStyleComboBox->currentIndex()==0)
	{
		symbolSizeDoubleSpinBox->setEnabled(false);
	}
	else
	{
		symbolSizeDoubleSpinBox->setEnabled(true);
	}
	if (lineAlignmentComboBox->currentIndex()==0)
		plotCurve->setCurveType(QwtPlotCurve::Yfx);
	else
		plotCurve->setCurveType(QwtPlotCurve::Xfy);

	histogramItem->setHistogramAttribute(HistogramItem::Xfy, histogramAlignmentComboBox->currentIndex()!=0);
	histogramItem->setColor(colorComboBox->currentColor());
}
void QwtWindowTab::on_colorComboBox_activated(QColor color)
{
	if (tabWidget->currentIndex()==LINE_INDEX || tabWidget->currentIndex()==SYMBOL_INDEX)
	{
		BuildPlotCurveFromSettings();
	}
	else if (tabWidget->currentIndex()==HISTOGRAM_INDEX)
	{
		histogramItem->setColor(color);
	}
// 	else
// 	{
// 		BuildSymbolFromSettings();
// 	}

	qwtWindow->Replot();
}
// void QwtWindowTab::BuildSymbolFromSettings()
// {
// 	plotCurve->hide();
// 	plotCurve->detach();
// 	UpdateDataSource(false);
// }
void QwtWindowTab::on_timeColumnComboBox_activated(QString string)
{
	if (tabWidget->currentIndex()==LINE_INDEX || tabWidget->currentIndex()==SYMBOL_INDEX)
	{

	}
	else if (tabWidget->currentIndex()==HISTOGRAM_INDEX)
	{
	}

	UpdateDataSource(true);
	qwtWindow->Replot();
}
void QwtWindowTab::RebuildOnAnimationTick(void)
{
//	DetachCurveList();
	int idx = tabWidget->currentIndex();
//	if (idx==LINE_INDEX || idx==HISTOGRAM_INDEX)
//	{
		UpdateDataSource(false);
// 	}
// 	else
// 	{
// 		BuildSymbolFromSettings();
// 	}
}
void QwtWindowTab::UpdateBasedOnTabWidget(void)
{
//	DetachCurveList();
	if (GetAMC()->GetQuery().isEmpty())
		return;

	int idx = tabWidget->currentIndex();
	if (idx==LINE_INDEX || idx==SYMBOL_INDEX)
	{
		plotCurve->detach();
		plotCurve->attach(qwtPlot);
		plotCurve->show();
		BuildPlotCurveFromSettings();
		UpdateDataSource(false);

//		plotCurve->attach(qwtPlot);
		histogramItem->detach();
		histogramItem->hide();
	}
	else if (idx==HISTOGRAM_INDEX)
	{
		plotCurve->detach();
		plotCurve->hide();
	//	plotCurve->detach();
		BuildPlotCurveFromSettings();
		UpdateDataSource(false);
		histogramItem->detach();
		histogramItem->attach(qwtPlot);
		histogramItem->show();
	}
// 	else
// 	{
// 		plotCurve->hide();
// 		plotCurve->detach();
// 		histogramItem->hide();
// 		histogramItem->detach();
// 		BuildSymbolFromSettings();
// 	}

	qwtWindow->Replot();
}
void QwtWindowTab::on_tabWidget_currentChanged(int idx)
{
	UpdateBasedOnTabWidget();
};
void QwtWindowTab::on_curveStyleComboBox_activated(int index)
{
	plotCurve->setStyle((QwtPlotCurve::CurveStyle)curveStyleComboBox->currentIndex());	
	plotCurve->setPen(*pen);
	qwtWindow->Replot();

	if (index==0)
	{
		penStyleComboBox->setEnabled(false);
		linePenWidthSpinBox->setEnabled(false);
	}
	else
	{
		penStyleComboBox->setEnabled(true);
		linePenWidthSpinBox->setEnabled(true);
	}
}
void QwtWindowTab::on_penStyleComboBox_activated(QString string)
{
	pen->setStyle((Qt::PenStyle) penStyleComboBox->currentIndex());
	plotCurve->setPen(*pen);
	qwtWindow->Replot();
}
void QwtWindowTab::on_linePenWidthSpinBox_valueChanged(int newValue)
{
	pen->setWidth(linePenWidthSpinBox->value());
	plotCurve->setPen(*pen);
	qwtWindow->Replot();
}
void QwtWindowTab::on_symbolPenWidthSpinBox_valueChanged(int newValue)
{
	qwtWindow->Replot();
}
void QwtWindowTab::on_baselineDoubleSpinBox_valueChanged(double newValue)
{
	histogramItem->setBaseline(newValue);
	qwtWindow->Replot();
}
void QwtWindowTab::on_symbolStyleComboBox_activated(int index)
{
	symbol->setStyle((QwtSymbol::Style)(symbolStyleComboBox->currentIndex()-1));

	plotCurve->setSymbol(*symbol);

	qwtWindow->Replot();

	if (index==0)
	{
		symbolSizeDoubleSpinBox->setEnabled(false);
	}
	else
	{
		symbolSizeDoubleSpinBox->setEnabled(true);
	}
}
void QwtWindowTab::on_symbolStyleComboBox_2_activated(int index)
{
//	BuildSymbolFromSettings();
//	qwtWindow->Replot();

	symbol->setStyle((QwtSymbol::Style)(symbolStyleComboBox_2->currentIndex()));
	plotCurve->setSymbol(*symbol);
	qwtWindow->Replot();
}
void QwtWindowTab::on_symbolSizeDoubleSpinBox_valueChanged(double newValue)
{
	symbol->setSize(newValue, -1);
	plotCurve->setSymbol(*symbol);

	qwtWindow->Replot();
}
void QwtWindowTab::UpdateCurveTitle(void)
{
	if (qwtWindow->HasDefaultTabName(this))
	{
		if (histogramAlignmentComboBox->currentIndex()==0)
		{
			// y = f(x)
			plotCurve->setTitle(yColumnComboBox->currentText() + QString("=f(") + xColumnComboBox->currentText() + QString(")"));
		}
		else
		{
			// x = f(y)
			plotCurve->setTitle(xColumnComboBox->currentText() + QString("=f(") + yColumnComboBox->currentText() + QString(")"));
		}
	}
	else
		plotCurve->setTitle(qwtWindow->GetTabName(this));

	histogramItem->setTitle(plotCurve->title());

}
void QwtWindowTab::UpdateDataSource(bool useDivisionCount)
{
	inUpdateDataSource=true;

	int rowStart, rowEnd;
	rowStart=0;
	rowEnd=GetModel()->constrainedRowCount(99999999);

	UpdateCurveTitle();

	SymbolSizeAlgorithm algorithm=(SymbolSizeAlgorithm) symbolSizeAlgorithmComboBox->currentIndex();
	QwtArray<double> xData;
	QwtArray<double> yData;
	QwtArray<double> magnitudeData;
	int xColumn = EchoChamber::GetColumnIndex(GetModel(), xColumnComboBox->currentText());
	int yColumn = EchoChamber::GetColumnIndex(GetModel(), yColumnComboBox->currentText());
	int magnitudeColumn = EchoChamber::GetColumnIndex(GetModel(), symbolMagnitudeComboBox->currentText());

	// Histogram interval is max-min / 10
	double minX=DBL_MAX,maxX=DBL_MIN;
	double minMagnitude=DBL_MAX,maxMagnitude=DBL_MIN;

	if (xColumn!=-1 && yColumn!=-1)
	{
		for (int row = rowStart; row < rowEnd; ++row)
		{
			QModelIndex xIndex = GetModel()->index(row, xColumn);
			QModelIndex yIndex = GetModel()->index(row, yColumn);
			QModelIndex magnitudeIndex = GetModel()->index(row, magnitudeColumn);
			QVariant xValueVariant = GetModel()->data(xIndex);
			QVariant yValueVariant = GetModel()->data(yIndex);
			QVariant magnitudeValueVariant = GetModel()->data(magnitudeIndex);

			if (xValueVariant.isNull()==false && yValueVariant.isNull()==false)
			{
				double xValue = xValueVariant.toDouble();
				double yValue = yValueVariant.toDouble();
				xData.push_back(xValue);
				yData.push_back(yValue);

				if (xValue<minX)
					minX=xValue;
				if (xValue>maxX)
					maxX=xValue;
			}

			if (magnitudeValueVariant.isNull()==false)
			{
				double magnitudeValue = magnitudeValueVariant.toDouble();
				if (magnitudeValue>=0.0)
				{
					magnitudeData.push_back(magnitudeValue);

					if (magnitudeValue>maxMagnitude)
						maxMagnitude=magnitudeValue;
					if (magnitudeValue<minMagnitude)
						minMagnitude=magnitudeValue;
				}
				else
					magnitudeData.push_back(0.0);
			}
			else
				magnitudeData.push_back(0.0);
		}
	}

	if (maxX>minX)
	{
		double histogramInterval;
		int numDivisions;

		if (useDivisionCount)
		{
			numDivisions = divisionCountSpinBox->value();
			histogramInterval=(maxX-minX)/numDivisions;
			histogramIntervalDoubleSpinBox->setValue(histogramInterval);
			histogramIntervalDoubleSpinBox->setSingleStep(histogramInterval/10.0);
		}
		else
		{
			int oldDivisionCount = divisionCountSpinBox->value();
			histogramInterval=histogramIntervalDoubleSpinBox->value();
			numDivisions=(maxX-minX)/histogramInterval;
			histogramInterval=(maxX-minX)/numDivisions;
			if (oldDivisionCount!=numDivisions)
				divisionCountSpinBox->setValue(numDivisions);
		}

		assert(numDivisions>0);

		QwtArray<QwtDoubleInterval> intervals;
		QwtArray<double> values;
		for (int i=0; i < numDivisions; i++)
		{
			intervals.push_back(QwtDoubleInterval(i*histogramInterval, (i+1)*histogramInterval));
			values.push_back(0.0);
		}

		// Insert into values
		for (int i=0; i < xData.size(); i++)
		{
			double xValue = xData[i];
			double yValue = yData[i];
			int slot = (xValue-minX)/histogramInterval;
			if (slot>=numDivisions)
				slot=numDivisions-1;
			values[slot]+=yValue;
		}

		// Then, define the intervals
		histogramItem->setData(QwtIntervalData(intervals,values));
	}


	bool useAbsolute = absoluteBaselineCheckbox->isChecked();
	double baseline;
	if (useAbsolute)
	{
		baseline=baselineSymbolDoubleSpinBox->value();
	}
	else
	{
		baseline=minMagnitude+baselineSymbolDoubleSpinBox->value();
	}
	maxMagnitude-=baseline;
	if (tabWidget->currentIndex()==SYMBOL_INDEX && maxMagnitude>0)
	{
		maxMagnitude=ApplyAlgorithm(algorithm, maxMagnitude);

		for (int i=0; i < magnitudeData.size(); i++)
		{
			double size;
			double normalizedValue = magnitudeData[i];
			normalizedValue-=baseline;
			normalizedValue=ApplyAlgorithm(algorithm, normalizedValue);
			size=normalizedValue;
			magnitudeData[i]=size;
		}

		plotCurve->setData3(xData,yData,magnitudeData);
	}
	else
		plotCurve->setData2(xData,yData);

	// Force axes to update
	if (qwtPlot->axisAutoScale(0))
		qwtPlot->setAxisAutoScale(0);
	if (qwtPlot->axisAutoScale(1))
		qwtPlot->setAxisAutoScale(1);

	inUpdateDataSource=false;

		/*
	}
	else
	{
		DetachCurveList();

		SymbolSizeAlgorithm algorithm=(SymbolSizeAlgorithm) symbolSizeAlgorithmComboBox->currentIndex();
		
		QwtArray<double> xData;
		QwtArray<double> yData;
		QwtArray<double> magnitudeData;
		int xColumn = EchoChamber::GetColumnIndex(GetModel(), xColumnComboBox->currentText());
		int yColumn = EchoChamber::GetColumnIndex(GetModel(), yColumnComboBox->currentText());
		int magnitudeColumn = EchoChamber::GetColumnIndex(GetModel(), symbolMagnitudeComboBox->currentText());
		double min=DBL_MAX,max=DBL_MIN;

		if (xColumn!=-1 && yColumn!=-1 && magnitudeColumn!=-1)
		{
			for (int row = rowStart; row < rowEnd; ++row)
			{
				QModelIndex xIndex = GetModel()->index(row, xColumn);
				QModelIndex yIndex = GetModel()->index(row, yColumn);
				QModelIndex magnitudeIndex = GetModel()->index(row, magnitudeColumn);
				QVariant xValueVariant = GetModel()->data(xIndex);
				QVariant yValueVariant = GetModel()->data(yIndex);
				QVariant magnitudeValueVariant = GetModel()->data(magnitudeIndex);

				if (xValueVariant.isNull()==false &&
					yValueVariant.isNull()==false &&
					magnitudeValueVariant.isNull()==false)
				{
					double xValue = xValueVariant.toDouble();
					double yValue = yValueVariant.toDouble();
					double magnitudeValue = magnitudeValueVariant.toDouble();
					if (magnitudeValue>0.0)
					{
						xData.push_back(xValue);
						yData.push_back(yValue);
						magnitudeData.push_back(magnitudeValue);

						if (magnitudeValue>max)
							max=magnitudeValue;
						if (magnitudeValue<min)
							min=magnitudeValue;
					}
				}
			}
		}

		if (xData.size()==0)
		{
			inUpdateDataSource=false;
			return;
		}

//		for (int i=0; i < xData.size(); i++)
//		{
//			plotList.push_back(new QwtPlotCurve());
//		}

		QwtSymbol sym;
		sym.setPen(QColor(Qt::black));
		sym.setStyle((QwtSymbol::Style)(symbolStyleComboBox_2->currentIndex()));
		sym.setBrush(colorComboBox->currentColor());

		baselineSymbolDoubleSpinBox->setSingleStep((max-min)/10.0);
		bool useAbsolute = absoluteBaselineCheckbox->isChecked();
		double baseline;
		if (useAbsolute)
		{
			baseline=baselineSymbolDoubleSpinBox->value();
		}
		else
		{
			baseline=min+baselineSymbolDoubleSpinBox->value();
		}


		max-=baseline;
		if (max<=0)
		{
			inUpdateDataSource=false;
			return;
		}
		max=ApplyAlgorithm(algorithm, max);
		
		for (int i=0; i < xData.size(); i++)
		{
			double size;
			double normalizedValue = magnitudeData[i];
			normalizedValue-=baseline;
			normalizedValue=ApplyAlgorithm(algorithm, normalizedValue);
			
			size=normalizedValue;

			sym.setSize(size);
			plotList[i]->setStyle(QwtPlotCurve::NoCurve);
			plotList[i]->setSymbol(sym);
			plotList[i]->setData(&xData[i], &yData[i], 1);
			plotList[i]->attach(qwtPlot);
			plotList[i]->show();
		}
	}
	*/


	inUpdateDataSource=false;
}
/*
void QwtWindowTab::DetachCurveList(void)
{
	int i;
	for (i=0; i < plotList.size(); i++)
	{
		plotList[i]->hide();
		plotList[i]->detach();
	}
	plotList.clear();
	// symbolList.clear();
}
*/
void QwtWindowTab::on_histogramIntervalDoubleSpinBox_valueChanged(double newValue)
{
	if (inUpdateDataSource==false)
	{
		UpdateDataSource(false);
		qwtWindow->Replot();
	}
}
void QwtWindowTab::on_divisionCountSpinBox_valueChanged(int newValue)
{
	if (inUpdateDataSource==false)
	{
		UpdateDataSource(true);
		qwtWindow->Replot();
	}
}
void QwtWindowTab::on_histogramAlignmentComboBox_activated(int index)
{
	if (inUpdateDataSource==false)
	{
		histogramItem->setHistogramAttribute(HistogramItem::Xfy, index!=0);
		qwtWindow->Replot();
	}
}
void QwtWindowTab::on_lineAlignmentComboBox_activated(int index)
{
	if (lineAlignmentComboBox->currentIndex()==0)
		plotCurve->setCurveType(QwtPlotCurve::Yfx);
	else
		plotCurve->setCurveType(QwtPlotCurve::Xfy);

	qwtWindow->Replot();
}
void QwtWindowTab::on_symbolMagnitudeComboBox_activated(int index)
{
	UpdateDataSource(false);
	qwtWindow->Replot();
}
void QwtWindowTab::on_symbolSizeAlgorithmComboBox_activated(int index)
{
	UpdateDataSource(false);
	qwtWindow->Replot();
}
void QwtWindowTab::on_maxSymbolPixelsSpinBox_valueChanged(int newValue)
{
	if (inUpdateDataSource==false)
	{
		UpdateDataSource(false);
		qwtWindow->Replot();
	}
}
void QwtWindowTab::on_baselineSymbolDoubleSpinBox_valueChanged(double newValue)
{
	if (inUpdateDataSource==false)
	{
		UpdateDataSource(false);
		qwtWindow->Replot();
	}
}
void QwtWindowTab::on_absoluteBaselineCheckbox_toggled(bool isSet)
{
	if (inUpdateDataSource==false)
	{
		UpdateDataSource(false);
		qwtWindow->Replot();
	}
}

void QwtWindowTab::Save(XMLSaver &xmlSaver)
{
	QByteArray geo_data;
	geo_data = saveGeometry();
	int versionNumber=0;

	AMC_Save(xmlSaver);
	xmlSaver.SaveElement("DBName", GetDBName());
	xmlSaver.SaveElement("Query", GetQuery());

	XML_SAVER_SAVE_ELEMENT(xmlSaver, geo_data);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, versionNumber);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, xColumnComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, yColumnComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, colorComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, tabWidget);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, curveStyleComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, penStyleComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, symbolStyleComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, linePenWidthSpinBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, symbolSizeDoubleSpinBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, baselineDoubleSpinBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, histogramAlignmentComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, histogramIntervalDoubleSpinBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, divisionCountSpinBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, symbolStyleComboBox_2);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, absoluteBaselineCheckbox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, symbolMagnitudeComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, baselineSymbolDoubleSpinBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, symbolSizeAlgorithmComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, lineAlignmentComboBox);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, histogramAlignmentComboBox);
	xmlSaver.SaveElement("TabName", qwtWindow->GetTabName(this));
}
void QwtWindowTab::Load(XMLSaver &xmlSaver)
{
	QByteArray geo_data;
	int versionNumber;
	AMC_Load(xmlSaver);
	QString database, query;
	xmlSaver.LoadElement("DBName", database, "");
	xmlSaver.LoadElement("Query", query, "");
	PostDBInit(query, database);

	XML_SAVER_LOAD_ELEMENT(xmlSaver, geo_data, false);
	restoreGeometry(geo_data);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, versionNumber, "-1");
	assert(versionNumber==0);

	XML_SAVER_LOAD_ELEMENT(xmlSaver, xColumnComboBox, "");
 	XML_SAVER_LOAD_ELEMENT(xmlSaver, yColumnComboBox, "");
 	XML_SAVER_LOAD_ELEMENT(xmlSaver, colorComboBox, true);
 	XML_SAVER_LOAD_ELEMENT(xmlSaver, tabWidget, 0);
 	XML_SAVER_LOAD_ELEMENT(xmlSaver, curveStyleComboBox, "");
 	XML_SAVER_LOAD_ELEMENT(xmlSaver, penStyleComboBox, "");
 	XML_SAVER_LOAD_ELEMENT(xmlSaver, symbolStyleComboBox, true);
 	XML_SAVER_LOAD_ELEMENT(xmlSaver, linePenWidthSpinBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, symbolSizeDoubleSpinBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, baselineDoubleSpinBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, histogramAlignmentComboBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, histogramIntervalDoubleSpinBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, divisionCountSpinBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, symbolStyleComboBox_2, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, absoluteBaselineCheckbox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, symbolMagnitudeComboBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, baselineSymbolDoubleSpinBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, symbolSizeAlgorithmComboBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, lineAlignmentComboBox, 0);
	XML_SAVER_LOAD_ELEMENT(xmlSaver, histogramAlignmentComboBox, 0);
	QString tabName;
	xmlSaver.LoadElement("TabName", tabName, "");
	if (tabName.isEmpty()==false)
		qwtWindow->RenameTab(tabName, this);

//	DetachCurveList();
	UpdateBasedOnTabWidget();
	UpdateCurveTitle();
}
void QwtWindowTab::on_windowTabLineEdit_editingFinished()
{
	qwtWindow->RenameTab(windowTabLineEdit->text(), this);
	UpdateCurveTitle();
}
bool QwtWindowTab::HasPlotItem(QwtPlotItem *pi) const
{
	return plotCurve==pi || histogramItem==pi;
}
void QwtWindowTab::SetAnimationRowIndices(int from, int to, double valueBefore, double valueAfter)
{
	AnimatedModel::SetAnimationRowIndices(from,to,valueBefore,valueAfter);
	qwtWindow->SetAnimationRowIndices(from, to, valueBefore, valueAfter);
}

void QwtSymbolCurveLegendItem::drawIdentifier(QPainter *painter, const QRect &rect) const
{
	if (symbolCurve->IsUsingSymbolPixelSizes())
	{
		QSize symbolSize = 
			QwtPainter::metricsMap().screenToLayout(symbol().size());

		// scale the symbol size down if it doesn't fit into rect.

		if ( rect.width() < symbolSize.width() )
		{
			const double ratio = 
				double(symbolSize.width()) / double(rect.width());
			symbolSize.setWidth(rect.width());
			symbolSize.setHeight(qRound(symbolSize.height() / ratio));
		}
		if ( rect.height() < symbolSize.height() )
		{
			const double ratio = 
				double(symbolSize.width()) / double(rect.width());
			symbolSize.setHeight(rect.height());
			symbolSize.setWidth(qRound(symbolSize.width() / ratio));
		}

		QRect symbolRect;
		symbolRect.setSize(symbolSize);
		symbolRect.moveCenter(rect.center());

		painter->save();
		painter->setBrush(symbol().brush());
		painter->setPen(QwtPainter::scaledPen(symbol().pen()));
		symbol().draw(painter, symbolRect);
		painter->restore();
	}
	else
	{
		QwtLegendItem::drawIdentifier(painter, rect);
	}
}
class PixelMatrix : private QBitArray
{
public:
	PixelMatrix(const QRect& rect):
	  QBitArray(rect.width() * rect.height()),
		  _rect(rect)
	  {
		  fill(false);
	  }

	  inline bool testPixel(const QPoint& pos)
	  {
		  if ( !_rect.contains(pos) )
			  return false;

		  const int idx = _rect.width() * (pos.y() - _rect.y()) + 
			  (pos.x() - _rect.x());

		  const bool marked = testBit(idx);
		  if ( !marked )
			  setBit(idx, true);

		  return !marked;
	  }

private:
	QRect _rect;
};
QWidget *QwtSymbolCurve::legendItem() const
{
	return new QwtSymbolCurveLegendItem(this);
}
void QwtSymbolCurve::setData2(const QwtArray<double> &xData, const QwtArray<double> &yData)
{
	QwtPlotCurve::setData(xData,yData);
	symbolPixelSizes.clear();
}
void QwtSymbolCurve::setData3(const QwtArray<double> &xData, const QwtArray<double> &yData, const QwtArray<double> &sizeData)
{
	QwtPlotCurve::setData(xData,yData);
	symbolPixelSizes=sizeData;
}
bool QwtSymbolCurve::IsUsingSymbolPixelSizes(void) const
{
	return symbolPixelSizes.size()>0;
}
void QwtSymbolCurve::drawSymbols(QPainter *painter, const QwtSymbol &symbol,
						 const QwtScaleMap &xMap, const QwtScaleMap &yMap,
						 int from, int to) const
{
	if (IsUsingSymbolPixelSizes())
	{
		painter->setBrush(symbol.brush());
		painter->setPen(QwtPainter::scaledPen(symbol.pen()));

		const QwtMetricsMap &metricsMap = QwtPainter::metricsMap();

		QRect rect;

		
		if ( to > from && testPaintAttribute(PaintFiltered) )
		{
			const QRect window = painter->window();
			if ( window.isEmpty() )
				return;

			PixelMatrix pixelMatrix(window);

			for (int i = from; i <= to; i++)
			{
				const QPoint pi( xMap.transform(x(i)),
					yMap.transform(y(i)) );

				if ( pixelMatrix.testPixel(pi) )
				{
					rect.setSize(metricsMap.screenToLayout(QSize(symbolPixelSizes[i],symbolPixelSizes[i])));
					rect.moveCenter(pi);
					symbol.draw(painter, rect);
				}
			}
		}
		else
		{
			for (int i = from; i <= to; i++)
			{
				const int xi = xMap.transform(x(i));
				const int yi = yMap.transform(y(i));

				rect.setSize(metricsMap.screenToLayout(QSize(symbolPixelSizes[i],symbolPixelSizes[i])));
				rect.moveCenter(QPoint(xi, yi));
				symbol.draw(painter, rect);
			}
		}
	}
	else
	{
		QwtPlotCurve::drawSymbols(painter, symbol, xMap, yMap, from, to);
	}
}
