#ifndef SCATTERPLOTPOINTTAB_H
#define SCATTERPLOTPOINTTAB_H

// GLEW needs to be 1st include where ever OpenGL is used.
#include <GL/glew.h>

#include <QWidget>

#include "ui_ScatterPlotPointTab.h"
#include "AnimatedModelContainer.h"
#include "QDataTypes3D.h"

using namespace Ui;

class ScatterPlotWindow;
class ScatterPlotPointTab;

#define MAX_POINTS 2048

enum ScatterPlotPointTabColorMode
{
    SPPTCM_COLOR_BY_POSITION = 0,
    SPPTCM_USE_CONSISTENT_FIXED_COLOR,
    SPPTCM_USE_PER_POINT_COLOR,
    SPPTCM_USE_CONSISTENT_HEAT_MAP_INTENSITY,
    SPPTCM_USE_PER_POINT_HEAT_MAP_INTENSITY,
};

struct ColoredPoint
{
    QPoint3F point;
    // Used with SPPTCM_USE_PER_POINT_COLOR and SPPTCM_USE_CONSISTENT_FIXED_COLOR
    unsigned char perPointColor[4];
    // Used with SPPTCM_USE_PER_POINT_HEAT_MAP_INTENSITY and SPPTCM_USE_CONSISTENT_HEAT_MAP_INTENSITY
    float perPointIntensity;
};

struct ScatterPlotFrameInput
{
    // List of points to draw
    ColoredPoint points[MAX_POINTS];
    unsigned int numPoints;

    // How to determine color to use
    ScatterPlotPointTabColorMode colorMode;

    // Used by all
    // Size of point to render
    float pointRadius;
    // Used only with SPPTCM_COLOR_BY_POSITION
    // Color of point at lowest value of Axis aligned bounding box
    // Each component has a lower value than each component of pointPositionalMaxColor
    QColor pointPositionalMinColor;
    // Used only with SPPTCM_COLOR_BY_POSITION
    // Color of point at highest value of Axis aligned bounding box
    QColor pointPositionalMaxColor;
    // Used with SPPTCM_USE_CONSISTENT_HEAT_MAP_INTENSITY and SPPTCM_USE_PER_POINT_HEAT_MAP_INTENSITY
    // How fast the color bleeds off from heatMapColorByIntensity to 0 over pointRadius
    // Use as exponential component
    float heatMapDiffusionRate;

    // Current texture handle (if selected) or 0 if none (0 replaces "useTextures=false")
    GLuint pointTextureId, heatMapColorTexture1, heatMapColorTexture2;
};

class ScatterPlotPointTab : public QWidget, public ScatterPlotPointTabClass, public AnimatedModel
{
    Q_OBJECT

public:
    ScatterPlotPointTab(QWidget* parent, Qt::WindowFlags f, ScatterPlotWindow *_scatterPlotWindow, QString query, QString database);
    virtual ~ScatterPlotPointTab();

    AnimatedModel *GetAMC(void) {return this;}

    void RebuildAll(void);

    // If you need input this tick, call this to calculate
    void CalculateInputThisFrame(void);

    // Returns what was calculated by CalculateInputThisFrame. Saved until next call to CalculateInputThisFrame
	ScatterPlotFrameInput * GetInputThisFrame(void) {return &inputThisFrame;}
	inline GLuint GetConsistentHeatMapTexture(void) const {return heatMapColorTexture1;}
	inline GLuint GetPerPointHeatMapTexture(void) const {return heatMapColorTexture2;}
	ScatterPlotPointTabColorMode GetColorMode(void) const {
        switch (tabWidget->currentIndex())
        {
            case 0: return SPPTCM_COLOR_BY_POSITION;
            case 1: return SPPTCM_USE_CONSISTENT_FIXED_COLOR;
            case 2: return SPPTCM_USE_PER_POINT_COLOR;
            case 3: return SPPTCM_USE_CONSISTENT_HEAT_MAP_INTENSITY;
            case 4: return SPPTCM_USE_PER_POINT_HEAT_MAP_INTENSITY;
        }
        return SPPTCM_COLOR_BY_POSITION;
    }

    // Bounding box around all points
    // Call AFTER CalculateInputThisFrame(), which calculates it

    inline const QAxisAlignedBoundingBox& GetBoundingBox(void) const
    {
        return boundingBox;
    }
    inline unsigned char calculateColorPortion(const float portion) const
    {
        float value = 255.0f * portion;
        if (value < 0.0f) return 0;
        else if (value > 255.0f) return 255;
        else return (unsigned char)value;
    }
    void SetPointTexture(QString fileName);
    void ChangeHeatMapTexture1(QString fileName);
    void ChangeHeatMapTexture2(QString fileName);
    void ClearPointTexture(void);

    void freeOpenGLResources();

	virtual void Save(XMLSaver &xmlSaver);
	virtual void Load(XMLSaver &xmlSaver);

public slots:
    void on_setTextureButton_released();
    void on_changeTexture1PushButton_released();
    void on_changeTexture2PushButton_released();
    void on_clearTextureButton_released();
    void on_xColumnComboBox_currentIndexChanged(int newIndex);
    void on_yColumnComboBox_currentIndexChanged(int newIndex);
    void on_zColumnComboBox_currentIndexChanged(int newIndex);
    void on_tabWidget_currentChanged(int idx);

protected:
    void DisplayTextureLoadError(QString fileName);
    void DeleteAll(void);
    void PostDBInit(QString query, QString database);
    ScatterPlotWindow * const scatterPlotWindow;
    GLuint pointTextureId, heatMapColorTexture1, heatMapColorTexture2;

    // Bounding box around all points
    QAxisAlignedBoundingBox boundingBox;
    bool calculateBoundingBox;
    ScatterPlotFrameInput inputThisFrame;
    QString currentPointTextureFilename, heatMapColorTexture1Filename, heatMapColorTexture2Filename;


};

#endif // SCATTERPLOTPOINTTAB_H
