#include "ScatterPlotWindow.h"
#include "ScatterPlotPointTab.h"
#include <QFileDialog>
#include <QMouseEvent>
#include <assert.h>
#include <math.h>
#include <QInputDialog>
#include <QSqlError>
#include <QMessageBox>
#include <QEvent>
#include <assert.h>
#include "EchoChamber.h"
#include "PointSorter.h"
#include <QGLFramebufferObject>
#include <QtGui>
#include <QtOpenGL>
#include <QtCore/QFile>
#include <QtGui/QFont>
#include <QtCore/qglobal.h>

#include <string>

extern EchoChamber *echoChamber;

#define BUFFER_GROWTH_FACTOR 1.2f // Determines the multiplier for growing an array to minimize re-allocations.

#define FBO_WIDTH	512
#define FBO_HEIGHT	512

static bool g_glewInitialized = false;
static QFont g_fontAxisLabel("Times", 12, QFont::Bold);

bool checkOpenGL()
{
    GLenum error = ::glGetError();
    if (error == GL_NO_ERROR) return true;
    else
    {
        qCritical("OpenGL Error Detected: %s\n", gluErrorString(error));
        return false;
    }
}

bool validateFeature(bool feature, const char * featureName)
{
    if (feature) return true;
    else
    {
        QString message;
        QTextStream stream(&message);
        stream << "Your OpenGL implementation does not support \'" << featureName << "\'.\nThis feature is required for the Heat Map visualization.\nPlease consider upgrading your card or drivers.";
        QMessageBox::warning(0, "OpenGL Warning", message);
        return false;
    }
}

bool necessaryExtensionsSupported()
{
    if (!validateFeature(GLEW_ARB_vertex_shader, "Vertex Shaders")) return false;
    if (!validateFeature(GLEW_ARB_fragment_shader, "Fragment Shaders")) return false;
    if (!validateFeature(GLEW_ARB_shader_objects, "Shader Objects")) return false;
    if (!validateFeature(GLEW_EXT_framebuffer_object, "Framebuffer Objects")) return false;
    return true;
}

ScatterPlotWindow::ScatterPlotWindow(QWidget* parent, Qt::WindowFlags f)
: QMainWindow(parent, f),
scatterPlotWidget(new ScatterPlotWidget(this, f))
{
    //scatterPlotWindow->GetAnimatedModelAtIndex(0)->GetModel()->setQuery(_query);

    setupUi(this);


}

void ScatterPlotWindow::InitWithDb(QStringList queries, QStringList databaseNames, bool startInHeatMapMode)
{
    // 	for (int i=0; i < databaseNames.size(); i++)
    // 		echoChamber->InitAMCWithDB(GetAnimatedModelAtIndex(i), queries[i], databaseNames[i]);

    newTabIndex = 2;

    //	mainToolBar->setVisible(false);
    scatterPlotWidgetLayout->addWidget(scatterPlotWidget);


    //	dockWidget_background->hide();
    actionBackground->setChecked(true);
    dockWidget_Camera->hide();
    dockWidget_Planes->hide();
    //	dockWidget->hide();
    actionPoints->setChecked(true);


    bool defaultWasSet;
    echoChamber->AddDefaultColors(backgroundColorCombobox);
    backgroundColorCombobox->setCurrentColor(Qt::white);
    QDir qDirHeatMapTextures("ScatterPlot/HeatMap", "*.png *.jpg *.bmp *.dds");

    for (int i = 0; i < databaseNames.size(); i++)
    {
        ScatterPlotPointTab *scatterPlotPointTab;
        if (i > 0)
            scatterPlotPointTab = AddTab(queries[i], databaseNames[i]);
        else
            scatterPlotPointTab = AddScatterPlotPointTabToNewestTabWidget(queries[i], databaseNames[i]);

        if (startInHeatMapMode)
        {
            scatterPlotPointTab->tabWidget->setCurrentIndex(SPPTCM_USE_CONSISTENT_HEAT_MAP_INTENSITY);
            scatterPlotPointTab->yColumnComboBox->setCurrentIndex(scatterPlotPointTab->yColumnComboBox->count() - 1);
            scatterPlotPointTab->pointRadiusSpinBox->setValue(12);
            QDir qDirPointTextures("ScatterPlot", "*.png *.jpg *.bmp *.dds");
            if (qDirPointTextures.count() > 0)
                scatterPlotPointTab->SetPointTexture(qDirPointTextures.absoluteFilePath(qDirPointTextures[0]));
            yAxisRadioButton->setChecked(true);
            getGLWidget()->on_yAxisRadioButton_released();
        }
        else
        {
            scatterPlotPointTab->tabWidget->setCurrentIndex(0);
        }

        if (qDirHeatMapTextures.count() > 0)
        {
            scatterPlotPointTab->ChangeHeatMapTexture1(qDirHeatMapTextures.absoluteFilePath(qDirHeatMapTextures[0]));
            scatterPlotPointTab->ChangeHeatMapTexture2(qDirHeatMapTextures.absoluteFilePath(qDirHeatMapTextures[0]));
        }
    }



    resize(512, 512);
    show();
    ResizeContents(this, centralWindowWidget);
}

ScatterPlotWindow::~ScatterPlotWindow()
{
    freeOpenGL();
    delete scatterPlotWidget;
}

void ScatterPlotWindow::on_addPlotPushButton_released()
{
    on_actionAdd_Tab_activated();
}

ScatterPlotPointTab * ScatterPlotWindow::AddTab(QString query, QString database)
{
    QWidget *tab = new QWidget();
    tab->setObjectName(QString::fromUtf8("tab"));
    QString tabName;
    tabName.sprintf("P%i", newTabIndex++);
    pointTabWidget->addTab(tab, tabName);
    return AddScatterPlotPointTabToNewestTabWidget(query, database);
}

void ScatterPlotWindow::on_actionAdd_Tab_activated()
{
    int i;
    for (i = 0; i < pointTabWidget->count(); i++)
    {
        if (windowTabs[i]->parent() == pointTabWidget->currentWidget())
        {
            AddTab(windowTabs[i]->GetAMC()->GetQuery(), windowTabs[i]->GetAMC()->GetDBName());
            return;
        }
    }
}

void ScatterPlotWindow::on_pointTabWidget_tabCloseRequested(int tabIndex)
{
    if (pointTabWidget->count() > 1)
    {
        bool wasRemoved = false;
        for (int i = 0; i < windowTabs.count(); i++)
        {
            if (windowTabs[i]->parent() == pointTabWidget->widget(tabIndex))
            {
                wasRemoved = true;
                windowTabs.removeAt(i);
                break;
            }
        }
        assert(wasRemoved == true);

        pointTabWidget->widget(tabIndex)->close();
        pointTabWidget->removeTab(tabIndex);
        // windowTabs.removeAt(tabIndex);

    }
}

ScatterPlotPointTab * ScatterPlotWindow::AddScatterPlotPointTabToNewestTabWidget(QString query, QString database)
{
    int newestWidgetIndex = pointTabWidget->count() - 1;
    QVBoxLayout *tabLayout = new QVBoxLayout(pointTabWidget->widget(newestWidgetIndex));
    pointTabWidget->widget(newestWidgetIndex)->setAttribute(Qt::WA_DeleteOnClose, true);
    tabLayout->setSpacing(0);
    tabLayout->setMargin(0);
    ScatterPlotPointTab *sppt = new ScatterPlotPointTab(pointTabWidget->widget(newestWidgetIndex), 0, this, query, database);
    sppt->setAttribute(Qt::WA_DeleteOnClose, true);
    tabLayout->addWidget(sppt);
    windowTabs.push_back(sppt);

    // Direct ScatterPlotPointTab calls to ScatterPlotWidget which uses them
    //connect(sppt->pointRadiusSpinBox, SIGNAL(valueChanged(int)), this, SLOT(on_ScatterPlotPointTab_pointRadiusSpinBox_valueChanged(int)));
    connect(sppt->pointRadiusSpinBox, SIGNAL(valueChanged(int)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_pointRadiusSpinBox_valueChanged(int)));
    connect(sppt->rateOfDiffusionSpinBox, SIGNAL(valueChanged(double)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_rateOfDiffusionSpinBox_valueChanged(double)));
    connect(sppt->xColumnComboBox, SIGNAL(currentIndexChanged(int)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_xColumnComboBox_currentIndexChanged(int)));
    connect(sppt->yColumnComboBox, SIGNAL(currentIndexChanged(int)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_yColumnComboBox_currentIndexChanged(int)));
    connect(sppt->zColumnComboBox, SIGNAL(currentIndexChanged(int)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_zColumnComboBox_currentIndexChanged(int)));
    connect(sppt->tabWidget, SIGNAL(currentChanged(int)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_tabWidget_currentChanged(int)));
    connect(sppt->minColorComboBox, SIGNAL(activated(QColor)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_minColorComboBox_activated(QColor)));
    connect(sppt->maxColorComboBox, SIGNAL(activated(QColor)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_maxColorComboBox_activated(QColor)));
    connect(sppt->fixedRGBColorComboBox, SIGNAL(activated(QColor)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_fixedRGBColorComboBox_activated(QColor)));
    connect(sppt->rgbByColumnComboBox, SIGNAL(currentIndexChanged(int)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_rgbByColumnComboBox_currentIndexChanged(int)));
    connect(sppt->intensityFixedSpinBox, SIGNAL(valueChanged(double)), scatterPlotWidget, SLOT(on_ScatterPlotPointTab_intensityFixedSpinBox_valueChanged(double)));

    // Direct ScatterPlotWindow calls to ScatterPlotWidget which uses them

    connect(backgroundButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_backgroundButton_released()));
    connect(clearBackgroundButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_clearBackgroundButton_released()));
    connect(showPlanesCheckBox, SIGNAL(toggled(bool)), scatterPlotWidget, SLOT(on_showPlanesCheckBox_toggled(bool)));
    connect(ticksPerPlaneSpinbox, SIGNAL(valueChanged(int)), scatterPlotWidget, SLOT(on_ticksPerPlaneSpinbox_valueChanged(int)));
    connect(saveAngleButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_saveAngleButton_released()));
    connect(xAxisRadioButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_xAxisRadioButton_released()));
    connect(yAxisRadioButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_yAxisRadioButton_released()));
    connect(zAxisRadioButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_zAxisRadioButton_released()));
    connect(perspectiveRadioButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_perspectiveRadioButton_released()));
    connect(userDefinedRadioButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_userDefinedRadioButton_released()));
    connect(userDefinedRadioButton, SIGNAL(released()), scatterPlotWidget, SLOT(on_userDefinedRadioButton_released()));
    connect(userDefinedCameraHeightSpinbox, SIGNAL(valueChanged(double)), scatterPlotWidget, SLOT(on_userDefinedCameraHeightSpinbox_valueChanged(double)));
    connect(userDefinedCameraAngleSpinbox, SIGNAL(valueChanged(double)), scatterPlotWidget, SLOT(on_userDefinedCameraAngleSpinbox_valueChanged(double)));

    return sppt;
}

void ScatterPlotWindow::on_actionEdit_SQL_triggered(bool b)
{
    QStringList stringListQueries, stringListDatabases;
    for (int i = 0; i < GetAMCCount(); i++)
    {
        stringListQueries.push_back(GetAnimatedModelAtIndex(i)->GetModel()->query().lastQuery());
        stringListDatabases.push_back(GetAnimatedModelAtIndex(i)->GetDBName());
    }
    echoChamber->SpawnSQLQueryTool(stringListQueries, stringListDatabases);
}

void ScatterPlotWindow::on_actionEdit_Tick_triggered(bool b)
{
    QString str = echoChamber->QueryForTickEdit(GetAnimatedModelAtIndex(0));
}

void ScatterPlotWindow::on_backgroundColorCombobox_activated(QColor color)
{
    getGLWidget()->SetBackgroundColor(color);
}

void ScatterPlotWindow::closeEvent(QCloseEvent * event)
{
    freeOpenGL();
}

void ScatterPlotWindow::freeOpenGL()
{
    for (int tabIndex = 0, tabCount = windowTabs.size(); tabIndex < tabCount; ++tabIndex)
    {
        ScatterPlotPointTab * tabCurrent = windowTabs[ tabIndex ];
        if (tabCurrent)
        {
            tabCurrent->freeOpenGLResources();
        }
    }
}

void ScatterPlotWidget::LoadBackground(QString fileName)
{
    QImage qImage;
    if (qImage.load(fileName))
    {
        createTexture(qImage, backgroundTexId);
        backgroundTextureFilename = fileName;
    }
}

void ScatterPlotWidget::ClearBackground(void)
{
    freeTexture(backgroundTexId);
    backgroundTextureFilename.clear();
}

void ScatterPlotWidget::on_backgroundButton_released()
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Image"), QString(), tr("Image Files (*.png *.jpg *.bmp *.dds)"));
    if (!fileName.isEmpty())
    {
        LoadBackground(fileName);
    }
}

void ScatterPlotWidget::on_clearBackgroundButton_released()
{
    ClearBackground();
}

void ScatterPlotWidget::on_showPlanesCheckBox_toggled(bool isSet) {
    //    showTicks = isSet;
}

void ScatterPlotWidget::on_ticksPerPlaneSpinbox_valueChanged(int curValue) {
    //   tickCount = (size_t)curValue;
}

void ScatterPlotWidget::on_saveAngleButton_released()
{
    ScatterPlotWindow *scatterPlotWindow = GetParent();
    spinBoxChangedInCode = true;
    scatterPlotWindow->userDefinedCameraHeightSpinbox->setValue(cameraCurrentHeight);
    scatterPlotWindow->userDefinedCameraAngleSpinbox->setValue(cameraCurrentAngle);
    spinBoxChangedInCode = false;
}

void ScatterPlotWidget::on_xAxisRadioButton_released()
{
    //cameraCurrentDistance = cameraDefaultDistance;
    cameraCurrentHeight = 0.0f;
    cameraCurrentAngle = -90.0f;
}

void ScatterPlotWidget::on_yAxisRadioButton_released()
{
    //cameraCurrentDistance = cameraDefaultDistance;
    cameraCurrentHeight = 90.0f;
    cameraCurrentAngle = 0.0f;
}

void ScatterPlotWidget::on_zAxisRadioButton_released()
{
    //cameraCurrentDistance = cameraDefaultDistance;
    cameraCurrentHeight = 0.0f;
    cameraCurrentAngle = 0.0f;
}

void ScatterPlotWidget::on_perspectiveRadioButton_released()
{
    cameraCurrentHeight = cameraDefaultHeight;
    cameraCurrentAngle = cameraDefaultAngle;
}

void ScatterPlotWidget::on_userDefinedRadioButton_released()
{
    ScatterPlotWindow *scatterPlotWindow = GetParent();
    cameraCurrentHeight = scatterPlotWindow->userDefinedCameraHeightSpinbox->value();
    cameraCurrentAngle = scatterPlotWindow->userDefinedCameraAngleSpinbox->value();
}

void ScatterPlotWidget::on_userDefinedCameraHeightSpinbox_valueChanged(double d)
{
    ScatterPlotWindow *scatterPlotWindow = GetParent();
    if (spinBoxChangedInCode == false)
        scatterPlotWindow->userDefinedRadioButton->setChecked(true);
    cameraCurrentHeight = d;
}

void ScatterPlotWidget::on_userDefinedCameraAngleSpinbox_valueChanged(double d)
{
    ScatterPlotWindow *scatterPlotWindow = GetParent();
    if (spinBoxChangedInCode == false)
        scatterPlotWindow->userDefinedRadioButton->setChecked(true);
    cameraCurrentAngle = d;
}
//void ScatterPlotWindow::on_ScatterPlotPointTab_pointRadiusSpinBox_valueChanged(int val)
//{
//	ScatterPlotPointTab *scatterPlotPointTab = (ScatterPlotPointTab*) pointTabWidget->currentWidget();
//}

void ScatterPlotWindow::Save(XMLSaver &xmlSaver)
{
    // this is always first
	SaveType(xmlSaver);
	SaveWindowTitle(xmlSaver,this);
	SaveVersion(xmlSaver);

    QByteArray geo_data = saveGeometry();
    QByteArray layout_data = saveState();
	XML_SAVER_SAVE_ELEMENT(xmlSaver, geo_data);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, layout_data);

	GetAnimatedModelAtIndex(0)->AMC_Save(xmlSaver);
	XML_SAVER_SAVE_ELEMENT(xmlSaver, backgroundColorCombobox);
	getGLWidget()->Save(xmlSaver);

	for (int i=0; i < windowTabs.count(); i++)
	{
		xmlSaver.CDDown(xmlSaver.MakeDirectory("WindowTabs"));
		windowTabs[i]->Save(xmlSaver);
		xmlSaver.CDUp();
	}
}

void ScatterPlotWindow::Load(XMLSaver &xmlSaver)
{
    int version;
    LoadVersion(xmlSaver, version);


    int i;
    for (i = 0; i < windowTabs.count(); i++)
        delete windowTabs[i];

    QByteArray geo_data;
    QByteArray layout_data;
    XML_SAVER_LOAD_ELEMENT(xmlSaver, geo_data, false);
    XML_SAVER_LOAD_ELEMENT(xmlSaver, layout_data, false);
    restoreGeometry(geo_data);
    restoreState(layout_data);

	XML_SAVER_LOAD_ELEMENT(xmlSaver, backgroundColorCombobox, false);
	getGLWidget()->SetBackgroundColor(backgroundColorCombobox->currentColor());
	getGLWidget()->Load(xmlSaver);

	pointTabWidget->clear();
	windowTabs.clear();

	if (xmlSaver.CDDown("WindowTabs"))
	{
		do
		{
			ScatterPlotPointTab * tab = AddTab(QString(), QString());
			tab->Load(xmlSaver);
		}  while (xmlSaver.CDNext("WindowTabs"));
		xmlSaver.CDUp();
	}


    actionPoints->setChecked(dockWidget->isVisible());
    actionBackground->setChecked(dockWidget_background->isVisible());
    actionPlanes->setChecked(dockWidget_Planes->isVisible());
    actionCamera->setChecked(dockWidget_Camera->isVisible());
}

QString ScatterPlotWindow::GetDBName(void) const {return windowTabs[0]->GetAMC()->GetDBName();}
AnimatedModel* ScatterPlotWindow::GetAnimatedModelAtIndex(int index) {return windowTabs[index]->GetAMC();}
int ScatterPlotWindow::GetAMCCount(void) const {return windowTabs.size();}

ScatterPlotWidget::ScatterPlotWidget(QWidget* parent, Qt::WindowFlags f) :
QGLWidget(parent, 0, f),
backgroundTexId(0),
tickId(0),
tickColor(0x000000cc),
tickLabelsSupported(false),
cameraDefaultDistance(3.0f),
cameraDefaultHeight(10.0f),
cameraDefaultAngle(-45.0f),
cameraCurrentDistance(cameraDefaultDistance),
cameraCurrentHeight(cameraDefaultHeight),
cameraCurrentAngle(cameraDefaultAngle),
cameraPreDragHeight(),
cameraPreDragAngle(),
cameraLogDistance(-1),
cameraPreDragLogDistance(),
pointCount(0),
pointArray(NULL),
pointCapacity(0),
pointColorArray(NULL),
pointColorCapacity(0),
lastPos(),
scatterPlotWindow(static_cast<ScatterPlotWindow *> (parent)),
spinBoxChangedInCode(false),
shaderProgram(0),
shaderVertexObject(0),
shaderFragmentObject(0),
intensitySamplerLocation(-1),
heatMapSamplerLocation(-1),
m_framebufferobject(NULL)
{
    // Validate that this Widget is only placed on an existing ScatterPlotWindow (ensures type-safety).
    assert(parent);

    connect(&animationTimer, SIGNAL(timeout()), this, SLOT(animationTimerUpdate()));
    setMouseTracking(true);
    setUpdatesEnabled(true);

    animationTimer.setInterval(30);
    animationTimer.setSingleShot(false);
    animationTimer.start(30);
}

ScatterPlotWidget::~ScatterPlotWidget()
{
    makeCurrent();
    if (isValid())
    {
        if (backgroundTexId)
        {
            deleteTexture(backgroundTexId);
            backgroundTexId = 0;
        }
        if (tickId)
        {
            ::glDeleteLists(tickId, 1);
            tickId = 0;
        }
        if (shaderVertexObject) ::glDeleteShader(shaderVertexObject);
        if (shaderFragmentObject) ::glDeleteShader(shaderFragmentObject);
        if (shaderProgram) ::glDeleteProgram(shaderProgram);
        if (m_framebufferobject) delete m_framebufferobject;
    }

    if (pointCapacity && pointArray)
    {
        delete[] pointArray;
        pointArray = NULL;
    }

    if (pointColorCapacity && pointColorArray)
    {
        delete[] pointColorArray;
        pointColorArray = NULL;
    }
}

GLuint ScatterPlotWidget::bindShader(GLenum shaderType, const char * pszShaderPath)
{
    assert(GLEW_VERSION_2_0);

    QFile fileShader(pszShaderPath);
    if (fileShader.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        std::string strFileSource;
        while (!fileShader.atEnd())
        {
            strFileSource.append(fileShader.readLine());
        }

        if (!strFileSource.empty())
        {
            GLuint shader = ::glCreateShader(shaderType);
            if (shader)
            {
                const char * pSources[ 1 ] = {strFileSource.c_str()};
                ::glShaderSource(shader, 1, pSources, NULL);
                ::glCompileShader(shader);

                GLint intSuccess;
                ::glGetShaderiv(shader, GL_COMPILE_STATUS, &intSuccess);

                if (intSuccess == GL_TRUE)
                {
                    return shader;
                }
                else
                {
                    GLchar message[256];
                    GLsizei messageLength;
                    ::glGetShaderInfoLog(shader, 256, &messageLength, message);
                    QMessageBox::warning(this, tr("Failed to link shaders"), QString(message));
                    ::glDeleteShader(shader);
                }
            }
        }
    }
    return 0;
}

bool ScatterPlotWidget::prepareShaders()
{
    assert(isValid()); // Ensure that shader preparation only occurs with a valid OpenGL context.

    if (shaderProgram) return true;
    else if (GLEW_VERSION_2_0)
    {
        GLchar message[256];
        GLsizei messageLength;

        shaderProgram = ::glCreateProgram();
        if (shaderProgram)
        {
            shaderVertexObject = bindShader(GL_VERTEX_SHADER, "Shaders/HeatMap.vsh");
            if (shaderVertexObject) ::glAttachShader(shaderProgram, shaderVertexObject);

            shaderFragmentObject = bindShader(GL_FRAGMENT_SHADER, "Shaders/HeatMap.fsh");
            if (shaderFragmentObject) ::glAttachShader(shaderProgram, shaderFragmentObject);

            ::glLinkProgram(shaderProgram);

            GLint intProgramStatus;
            ::glGetProgramiv(shaderProgram, GL_LINK_STATUS, &intProgramStatus);
            if (intProgramStatus == GL_TRUE)
            {
                // Bind Variables
                intensitySamplerLocation = ::glGetUniformLocation(shaderProgram, "intensitySampler");
                heatMapSamplerLocation = ::glGetUniformLocation(shaderProgram, "heatMapSampler");

                m_framebufferobject = new QGLFramebufferObject(FBO_WIDTH, FBO_HEIGHT, GL_TEXTURE_2D);

                return true;
            }
            else
            {
                ::glGetProgramInfoLog(shaderProgram, 256, &messageLength, message);
                QMessageBox::warning(this, tr("Failed to link shaders"), QString(message));
                ::glDeleteProgram(shaderProgram);
                shaderProgram = 0;
            }
        }
    }

    return false;
}

void ScatterPlotWidget::initializeGL()
{
    // Set background clear color
    QColor color = GetParent()->backgroundColorCombobox->currentColor();
    ::glClearColor(color.red(), color.green(), color.blue(), 1.0);

    // Set point and line smoothing
    ::glEnable(GL_POINT_SMOOTH);
    ::glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
    ::glPointSize(5.0f);

    // Point sprite setup.
    if (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_1_5)
    {
        ::glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE);
    }

    tickLabelsSupported = (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_2_0) ? true : false;

    if (!g_glewInitialized)
    {
        GLuint glewError = glewInit();
        if (glewError == GLEW_OK)
        {
            g_glewInitialized = true;
        }
        else return;
    }

    //        intensitySamplerLocation = shaderProgram->uniformLocation("intensitySampler");
    //        heatMapSamplerLocation = shaderProgram->uniformLocation("heatMapSampler");
    //        if (intensitySamplerLocation == -1)
    //            QMessageBox::warning(this, tr("Shaders/HeatMap.fsh"), "Can't find intensitySampler");
    //        else if (heatMapSamplerLocation == -1)
    //            QMessageBox::warning(this, tr("Shaders/HeatMap.fsh"), "Can't find heatMapSampler");
    //        if (intensitySamplerLocation == -1 || heatMapSamplerLocation == -1)
    //        {
    //            delete shaderProgram;
    //            shaderProgram = 0;
    //            delete heatMapShaderPS;
    //            heatMapShaderPS = 0;
    //            delete heatMapShaderVS;
    //            heatMapShaderVS = 0;
    //            return;
    //        }
    //    }
}

void ScatterPlotWidget::SetBackgroundColor(QColor color)
{
    makeCurrent();
    ::glClearColor(color.red(), color.green(), color.blue(), 0.0);
}

void qt_add_rect_to_array(const QRectF &r, float *array)
{
    qreal left = r.left();
    qreal right = r.right();
    qreal top = r.top();
    qreal bottom = r.bottom();

    array[0] = left;
    array[1] = top;
    array[2] = right;
    array[3] = top;
    array[4] = right;
    array[5] = bottom;
    array[6] = left;
    array[7] = bottom;
}

void ScatterPlotWidget::resizeGL(int w, int h)
{
    ::glViewport(0, 0, w, h);

    float fHalfWidth = (float) w * 0.5f;
    float fHalfHeight = (float) h * 0.5f;
    backgroundDims.setCoords(-fHalfWidth, fHalfHeight, fHalfWidth, -fHalfHeight);

    // Only update the camera perspective, position, and orientation if there is a renderable surface area.
    if (w && h)
    {
        setPerspective(60.0f, (float) w / (float) h, 0.1f, 10000.0f);
    }
}

static inline void DrawQuad()
{
    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.0);
    glVertex2f(-1.0, -1.0);
    glTexCoord2f(1.0, 0.0);
    glVertex2f(1.0, -1.0);
    glTexCoord2f(1.0, 1.0);
    glVertex2f(1.0, 1.0);
    glTexCoord2f(0.0, 1.0);
    glVertex2f(-1.0, 1.0);
    glEnd();
}

static bool DrawTexture(GLuint textureID, int windowWidth, int windowHeight)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(0, 0, windowWidth, windowHeight);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, textureID);
    glEnable(GL_TEXTURE_2D);

    glDisable(GL_DEPTH_TEST);
    glColor3f(1.0, 1.0, 1.0);

    DrawQuad();

    glDisable(GL_TEXTURE_2D);

    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    return true;
}

void ScatterPlotWidget::paintGL()
{
    ::glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    int w = this->width();
    int h = this->height();

    if ((w <= 0) || (h <= 0)) return;

    QAxisAlignedBoundingBox tickVolume(QPoint3F(-100.0f, -100.0f, -100.0f), QPoint3F(100.0f, 100.0f, 100.0f));

    ScatterPlotWindow * const scatterPlotWindow = GetParent();
    const int frameInputCount = scatterPlotWindow->GetPointSourceCount();
    if (frameInputCount)
    {
        tickVolume.reset();
        for (int pointSourceIndex = 0; pointSourceIndex < frameInputCount; ++pointSourceIndex)
        {
            scatterPlotWindow->CalculateInput(pointSourceIndex);
            tickVolume.add(scatterPlotWindow->GetBoundingBox(pointSourceIndex));
        }
    }
    const QPoint3F center = tickVolume.getCenter();

    if (cameraLogDistance < 0)
    {
        float minDistance = (tickVolume.getSquareWidth() < tickVolume.getSquareHeight()) ? tickVolume.getSquareHeight() : tickVolume.getSquareWidth();
        minDistance = (minDistance < tickVolume.getSquareDepth()) ? tickVolume.getSquareDepth() : minDistance;
        calculateInitialDistance(minDistance * 1.5f);
    }

    setPerspective(60.0f, (float) w / (float) h, cameraCurrentDistance * 0.001f, cameraCurrentDistance * 1000.0f);

    setCameraOrbit(center.getX(), center.getY(), center.getZ(), cameraCurrentDistance, cameraCurrentHeight, cameraCurrentAngle);

    // Extract the camera position and aim vector
    QMatrix4x4 cameraMatrix;
    ::glGetFloatv(GL_MODELVIEW_MATRIX, cameraMatrix.m);
    QMatrix4x4 cameraInvMatrix( cameraMatrix.inverse() );
    QPoint3F cameraPosition = cameraInvMatrix.transform( QPoint3F( 0.0f, 0.0f, 0.0f ) );
    QPoint3F cameraAimVector( center.getX() - cameraPosition.getX(), center.getY() - cameraPosition.getY(), center.getZ() - cameraPosition.getZ() );
    cameraAimVector.Normalize();

    int totalPoints = 0;
    for (int pointSourceIndex = 0, pointSourceCount = frameInputCount; pointSourceIndex < pointSourceCount; ++pointSourceIndex)
    {
        ScatterPlotFrameInput * const currentFrameInput = scatterPlotWindow->GetInputThisFrame(pointSourceIndex);
        totalPoints += currentFrameInput->numPoints;
    }

    PointSorter pointSorter(totalPoints, QPlane( cameraPosition, cameraAimVector ));
    for (int pointSourceIndex = 0, pointSourceCount = frameInputCount; pointSourceIndex < pointSourceCount; ++pointSourceIndex)
    {
        ScatterPlotFrameInput * const currentFrameInput = scatterPlotWindow->GetInputThisFrame(pointSourceIndex);
        for (int pointIndex = 0; pointIndex < currentFrameInput->numPoints; pointIndex++)
        {
            pointSorter.AddPoint( RenderablePoint( currentFrameInput->points[pointIndex].point, currentFrameInput->points[pointIndex].perPointColor, currentFrameInput ) );
        }
    }
    pointSorter.SortPoints();

    // Display the background texture (if present)
    if (backgroundTexId)
    {
        // Store the current projection matrix and set it the new projection to Orthographic.
        ::glMatrixMode(GL_PROJECTION);
        ::glPushMatrix();
        ::glLoadIdentity();
        ::gluOrtho2D(backgroundDims.left(), backgroundDims.right(), backgroundDims.bottom(), backgroundDims.top());
        ::glMatrixMode(GL_MODELVIEW);

        ::glPushMatrix();
        ::glLoadIdentity();

        ::glDisable(GL_DEPTH_TEST);

        ::glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
        drawTexture(backgroundDims, backgroundTexId, GL_TEXTURE_2D);
        ::glBindTexture(GL_TEXTURE_2D, 0);

        ::glPopMatrix();

        // Restore the pre-Orthographic projection matrix
        ::glMatrixMode(GL_PROJECTION);
        ::glPopMatrix();
        ::glMatrixMode(GL_MODELVIEW);
    }

    ::glEnable(GL_DEPTH_TEST);

    // Display the grid ticks (if present)
    if (GetParent()->showPlanesCheckBox->isChecked())
    {
        ::glDisable(GL_TEXTURE_2D);
        ::glBindTexture(GL_TEXTURE_2D, 0);
        ::glDisable(GL_POINT_SPRITE);
        ::glDisable(GL_BLEND);
        ::glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
        if (tickId) ::glCallList(tickId);
        else renderTicks(tickVolume);

        // Only render the tick labels on OpenGL implementations which support it.
        if (tickLabelsSupported)
        {
            ::glEnable(GL_TEXTURE_2D);
            renderLabels(tickVolume);
        }
    }

    ::glDisableClientState(GL_NORMAL_ARRAY);
    ::glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    ::glEnableClientState(GL_VERTEX_ARRAY);
    ::glEnableClientState(GL_COLOR_ARRAY);

    QRectF screenRect(0, 0, backgroundDims.width(), backgroundDims.height());

    for (int batchIndex = 0, batchCount = pointSorter.GetBatchCount(); batchIndex < batchCount; ++batchIndex)
    {
        RenderablePoint * const pRenderablePointBatch = pointSorter.GetBatch( batchIndex );

        assert( pRenderablePointBatch );

        ScatterPlotFrameInput * const pRenderablePointFrameInput = pRenderablePointBatch[ 0 ].frameInput;

        assert( pRenderablePointFrameInput );

        const unsigned int currentBatchCount = pointSorter.GetBatchSize( batchIndex );

        assert(currentBatchCount < MAX_POINTS); // Number of points exceeds buffer size.  Will ignore excess points in Release builds.

        if ((currentBatchCount > 0) && (currentBatchCount < MAX_POINTS))
        {
            setPointRenderStates( (GLfloat)pRenderablePointFrameInput->pointRadius, pRenderablePointFrameInput->pointTextureId );

            RenderablePoint & currentPoint = pRenderablePointBatch[ 0 ];

            switch (pRenderablePointFrameInput->colorMode)
            {
                case SPPTCM_COLOR_BY_POSITION:
                {
                    updateColorsByPosition( pRenderablePointBatch, currentBatchCount, tickVolume);
                }
                case SPPTCM_USE_CONSISTENT_FIXED_COLOR:
                case SPPTCM_USE_PER_POINT_COLOR:
                {
                    ::glColorPointer(4, GL_UNSIGNED_BYTE, sizeof ( RenderablePoint), &currentPoint.pointColor);
                    ::glVertexPointer(3, GL_FLOAT, sizeof ( RenderablePoint), &currentPoint.point.x);
                    ::glDrawArrays(GL_POINTS, 0, currentBatchCount);
                    break;
                }
                case SPPTCM_USE_CONSISTENT_HEAT_MAP_INTENSITY:
                case SPPTCM_USE_PER_POINT_HEAT_MAP_INTENSITY:
                {
                    if (!prepareShaders()) continue;

                    // Frame buffer object _must_ be ready at this point.
                    assert(m_framebufferobject);

                    GLuint heatmapSampler = (pRenderablePointFrameInput->colorMode == SPPTCM_USE_CONSISTENT_HEAT_MAP_INTENSITY) ? pRenderablePointFrameInput->heatMapColorTexture1 : pRenderablePointFrameInput->heatMapColorTexture2;
                    if (heatmapSampler == 0)
                        continue;

                    // Bind the render target
                    m_framebufferobject->bind();

                    // Push the viewport
                    glPushAttrib(GL_VIEWPORT_BIT);
                    glViewport(0, 0, FBO_WIDTH, FBO_HEIGHT);

                    // Clear render target to black
                    glPushAttrib(GL_COLOR_BUFFER_BIT);
                    glPushAttrib(GL_DEPTH_BUFFER_BIT);
                    ::glDisable(GL_DEPTH_TEST);
                    ::glClearColor(0, 0, 0, 0);
                    ::glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
                    glPopAttrib(); // GL_DEPTH_BUFFER_BIT
                    glPopAttrib(); // GL_COLOR_BUFFER_BIT

                    // Enable additive blending
                    ::glDisable(GL_ALPHA_TEST);
                    ::glEnable(GL_BLEND);
                    ::glBlendFunc(GL_ONE, GL_ONE);

                    // Set the color pointer
                    ::glColorPointer(4, GL_UNSIGNED_BYTE, sizeof ( RenderablePoint), &currentPoint.pointColor);
                    ::glVertexPointer(3, GL_FLOAT, sizeof ( RenderablePoint), &currentPoint.point.x);
                    ::glDrawArrays(GL_POINTS, 0, currentBatchCount);

                    ::glPopAttrib(); // GL_VIEWPORT_BIT

                    m_framebufferobject->release();

                    // Render the heatmap.

                    // Set the viewport
                    ::glPushAttrib(GL_VIEWPORT_BIT);
                    ::glViewport(0, 0, screenRect.width(), abs(screenRect.height()));

                    // Set projection matrix
                    ::glMatrixMode(GL_PROJECTION);
                    ::glPushMatrix();
                    ::glLoadIdentity();
                    ::glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

                    // Set model view matrix
                    ::glMatrixMode(GL_MODELVIEW);
                    ::glPushMatrix();
                    ::glLoadIdentity();

                    // Set sampler 1
                    // Has to be texture 0 second, not sure why
                    ::glActiveTexture(GL_TEXTURE0);
                    ::glBindTexture(GL_TEXTURE_2D, m_framebufferobject->texture());

                    // Set sampler 2
                    ::glActiveTexture(GL_TEXTURE1);
                    ::glBindTexture( GL_TEXTURE_2D, heatmapSampler );

                    ::glEnable(GL_TEXTURE_2D);
                    ::glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

                    if ((intensitySamplerLocation >= 0) && (heatMapSamplerLocation >= 0))
                    {
                        ::glUseProgram(shaderProgram);
                        ::glUniform1i(intensitySamplerLocation, 0);
                        ::glUniform1i(heatMapSamplerLocation, 1);
                    }
                    // Disable depth testing
                    //::glDisable(GL_DEPTH_TEST);
                    ::glDepthMask( GL_FALSE );

                    // Draw the quad
                    glColor4f( 0.0f, 1.0f, 0.0f, 1.0f );
                    DrawQuad();

                    // Disable the Shader Program
                    ::glUseProgram(0);

                    // Cleanup texture states, returning that active TEXTURE unit to 0.
                    ::glActiveTexture(GL_TEXTURE1);
                    ::glBindTexture(GL_TEXTURE_2D, 0);

                    ::glActiveTexture(GL_TEXTURE0);
                    ::glBindTexture(GL_TEXTURE_2D, 0);
                    
                    // Restore states
                    ::glDisable(GL_TEXTURE_2D);
                    ::glMatrixMode(GL_MODELVIEW);
                    ::glPopMatrix();
                    ::glMatrixMode(GL_PROJECTION);
                    ::glPopMatrix();
                    ::glPopAttrib();

                    break;
                }
            }

            assert(checkOpenGL());
        }
    }
    ::glDepthMask( GL_TRUE );

    ::glDisable(GL_ALPHA_TEST);
    ::glAlphaFunc(GL_ALWAYS, 1.0f);
    ::glDisableClientState(GL_COLOR_ARRAY);
    ::glDisableClientState(GL_VERTEX_ARRAY);

    GLenum error = ::glGetError();
    assert(checkOpenGL());
}

void ScatterPlotWidget::setPointRenderStates(GLfloat pointSize, GLuint textureId)
{
    ::glPointSize( pointSize );

    if (QGLFormat::openGLVersionFlags() & QGLFormat::OpenGL_Version_1_5)
    {
        ::glEnable(GL_BLEND);
        ::glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        if (textureId)
        {
            ::glEnable(GL_TEXTURE_2D);
            ::glBindTexture(GL_TEXTURE_2D, textureId);
            ::glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
            ::glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE);
            ::glEnable(GL_ALPHA_TEST);
            ::glAlphaFunc(GL_GEQUAL, 0.0039f);
        }
        else
        {
            ::glBindTexture(GL_TEXTURE_2D, 0);
            ::glDisable(GL_TEXTURE_2D);
            ::glDisable(GL_ALPHA_TEST);
            ::glAlphaFunc(GL_ALWAYS, 1.0f);
        }

        ::glEnable(GL_POINT_SPRITE);
    }
    else
    {
        ::glDisable(GL_TEXTURE_2D);
        ::glBindTexture(GL_TEXTURE_2D, 0);
        ::glDisable(GL_POINT_SPRITE);
        ::glDisable(GL_BLEND);
    }
}

void ScatterPlotWidget::mousePressEvent(QMouseEvent *event)
{
    lastPos = event->pos();
    cameraPreDragLogDistance = cameraLogDistance;
    cameraPreDragHeight = cameraCurrentHeight;
    cameraPreDragAngle = cameraCurrentAngle;
}

void ScatterPlotWidget::mouseDoubleClickEvent(QMouseEvent * event)
{
    if (event->buttons() & Qt::LeftButton)
    {
        cameraCurrentHeight = cameraDefaultHeight;
        cameraCurrentAngle = cameraDefaultAngle;
        cameraLogDistance = -1;
    }
}

void ScatterPlotWidget::mouseMoveEvent(QMouseEvent *event)
{
    int dx = event->x() - lastPos.x();
    int dy = event->y() - lastPos.y();

    if (event->buttons() & Qt::LeftButton)
    {
        cameraCurrentHeight = cameraPreDragHeight + ((float) dy * 0.5f);
        cameraCurrentAngle = cameraPreDragAngle + ((float) dx * 0.5f);
    }
    else if (event->buttons() & Qt::RightButton)
    {
        int logDistance = cameraPreDragLogDistance - dy;
        if (logDistance < 1) logDistance = 1;
        cameraCurrentDistance = logToDistance(logDistance);
        cameraLogDistance = logDistance;
        //cameraCurrentDistance = cameraPreDragDistance - ((float)dy * (cameraPreDragDistance * 0.02f));
    }
}

void ScatterPlotWidget::animationTimerUpdate()
{
    // Force redraw of GL every 30 ms
    updateGL();
}

void ScatterPlotWidget::on_ScatterPlotPointTab_pointRadiusSpinBox_valueChanged(int val)
{
    makeCurrent();
    if (isValid())
    {
        ::glPointSize((GLfloat) val);
    }
}

void ScatterPlotWidget::on_ScatterPlotPointTab_rateOfDiffusionSpinBox_valueChanged(double val) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_xColumnComboBox_currentIndexChanged(int idx) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_yColumnComboBox_currentIndexChanged(int idx) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_zColumnComboBox_currentIndexChanged(int idx) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_tabWidget_currentChanged(int idx) {
    // Pos, Fixed RGB, RGB by column, intensity, intensity by column
}

void ScatterPlotWidget::on_ScatterPlotPointTab_minColorComboBox_activated(QColor color) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_maxColorComboBox_activated(QColor color) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_fixedRGBColorComboBox_activated(QColor color) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_rgbByColumnComboBox_currentIndexChanged(int idx) { }

void ScatterPlotWidget::on_ScatterPlotPointTab_intensityFixedSpinBox_valueChanged(double val) { }

int ScatterPlotWidget::GetPointTabIndex(void) const
{
    return GetParent()->pointTabWidget->currentIndex();
}

void ScatterPlotWidget::setPerspective(float fovY, float aspectRatio, float nearClip, float farClip)
{
    assert(isValid());

    ::glMatrixMode(GL_PROJECTION);

    ::glLoadIdentity();
    ::gluPerspective(fovY, aspectRatio, nearClip, farClip);

    ::glMatrixMode(GL_MODELVIEW);
}

void ScatterPlotWidget::setCameraOrbit(GLfloat inAimX, GLfloat inAimY, GLfloat inAimZ, GLfloat inCamDistance, GLfloat inCamHeight, GLfloat inCamAngle)
{
    assert(isValid());

    ::glLoadIdentity();

    ::glTranslatef(0.0f, 0.0f, -inCamDistance);
    ::glRotatef(inCamHeight, 1.0f, 0.0f, 0.0f);
    ::glRotatef(inCamAngle, 0.0f, 1.0f, 0.0f);
    ::glTranslatef(-inAimX, -inAimY, -inAimZ);
}

void ScatterPlotWidget::renderTicks(const QAxisAlignedBoundingBox & tickVolume)
{
    assert(isValid());

    // Set the grid lines to the current tick color.
    ::glColor4bv((const GLbyte *) & tickColor);

    // If only 1 tick is requested, draw a simple x, y, z axis; otherwise draw the full ticked grid.
    int tickCount = GetParent()->ticksPerPlaneSpinbox->value();
    if (tickCount == 1)
    {
        ::glBegin(GL_LINE_STRIP);

        ::glVertex3f(tickVolume.getSquareMax().getX(), tickVolume.getSquareMin().getY(), tickVolume.getSquareMin().getZ());
        ::glVertex3f(tickVolume.getSquareMin().getX(), tickVolume.getSquareMin().getY(), tickVolume.getSquareMin().getZ());
        ::glVertex3f(tickVolume.getSquareMin().getX(), tickVolume.getSquareMax().getY(), tickVolume.getSquareMin().getZ());

        ::glEnd();

        ::glBegin(GL_LINES);

        ::glVertex3f(tickVolume.getSquareMin().getX(), tickVolume.getSquareMin().getY(), tickVolume.getSquareMin().getZ());
        ::glVertex3f(tickVolume.getSquareMin().getX(), tickVolume.getSquareMin().getY(), tickVolume.getSquareMax().getZ());

        ::glEnd();
    }
    else
    {
        const float invTickCount = 1.0f / (float) (tickCount - 1);

        const float fXSpacing = tickVolume.getSquareWidth() * invTickCount;
        const float fYSpacing = tickVolume.getSquareHeight() * invTickCount;
        const float fZSpacing = tickVolume.getSquareDepth() * invTickCount;

        ::glBegin(GL_LINES);

        // Build all 3 plane's grid-lines.
        for (size_t ii = 0; ii < tickCount; ++ii)
        {
            const float fXOffset = tickVolume.getSquareMin().getX() + ((float) ii * fXSpacing);
            const float fYOffset = tickVolume.getSquareMin().getY() + ((float) ii * fYSpacing);
            const float fZOffset = tickVolume.getSquareMin().getZ() + ((float) ii * fZSpacing);

            // Plane-lines along the Y axis.
            ::glVertex3f(tickVolume.getSquareMax().getX(), fYOffset, tickVolume.getSquareMin().getZ());
            ::glVertex3f(tickVolume.getSquareMin().getX(), fYOffset, tickVolume.getSquareMin().getZ());
            ::glVertex3f(tickVolume.getSquareMin().getX(), fYOffset, tickVolume.getSquareMin().getZ());
            ::glVertex3f(tickVolume.getSquareMin().getX(), fYOffset, tickVolume.getSquareMax().getZ());

            // Plane-lines along the Z axis.
            ::glVertex3f(tickVolume.getSquareMax().getX(), tickVolume.getSquareMin().getY(), fZOffset);
            ::glVertex3f(tickVolume.getSquareMin().getX(), tickVolume.getSquareMin().getY(), fZOffset);
            ::glVertex3f(tickVolume.getSquareMin().getX(), tickVolume.getSquareMin().getY(), fZOffset);
            ::glVertex3f(tickVolume.getSquareMin().getX(), tickVolume.getSquareMax().getY(), fZOffset);

            // Plane-lines along the X axis.
            ::glVertex3f(fXOffset, tickVolume.getSquareMax().getY(), tickVolume.getSquareMin().getZ());
            ::glVertex3f(fXOffset, tickVolume.getSquareMin().getY(), tickVolume.getSquareMin().getZ());
            ::glVertex3f(fXOffset, tickVolume.getSquareMin().getY(), tickVolume.getSquareMin().getZ());
            ::glVertex3f(fXOffset, tickVolume.getSquareMin().getY(), tickVolume.getSquareMax().getZ());
        }

        ::glEnd();
    }
}

void ScatterPlotWidget::createTicks(const QAxisAlignedBoundingBox & tickVolume)
{
    assert(isValid());

    if (tickId != 0)
    {
        ::glDeleteLists(tickId, 1);
        tickId = 0;
    }

    // Only create ticks when 1 or more are requested
    int tickCount = GetParent()->ticksPerPlaneSpinbox->value();
    if (tickCount > 0)
    {
        // Create a display list to contain the grid 'ticks'
        tickId = ::glGenLists(1);
        if (tickId)
        {
            ::glNewList(tickId, GL_COMPILE);

            renderTicks(tickVolume);

            ::glEndList();
        }
    }
}

void ScatterPlotWidget::renderLabels(const QAxisAlignedBoundingBox & tickVolume)
{
    assert(isValid());

    // Set the grid lines to the current tick color.
    qglColor(QColor(Qt::black));

    // Draw the axis labels.
    const float fExtentX = tickVolume.getSquareMax().getX() + (tickVolume.getWidth() * 0.01f);
    const float fExtentY = tickVolume.getSquareMax().getY() + (tickVolume.getHeight() * 0.01f);
    const float fExtentZ = tickVolume.getSquareMax().getZ() + (tickVolume.getDepth() * 0.01f);
    renderText(fExtentX, tickVolume.getSquareMin().getY(), tickVolume.getSquareMin().getZ(), QString("X"), g_fontAxisLabel);
    renderText(tickVolume.getSquareMin().getX(), fExtentY, tickVolume.getSquareMin().getZ(), QString("Y"), g_fontAxisLabel);
    renderText(tickVolume.getSquareMin().getX(), tickVolume.getSquareMin().getY(), fExtentZ, QString("Z"), g_fontAxisLabel);
    //    renderText(tickVolume.getSquareMax().getX(), tickVolume.getSquareMin().getY(), tickVolume.getSquareMin().getZ(), QString("X"), g_fontAxisLabel );
    //    renderText(tickVolume.getSquareMin().getX(), tickVolume.getSquareMax().getY(), tickVolume.getSquareMin().getZ(), QString("Y"), g_fontAxisLabel );
    //    renderText(tickVolume.getSquareMin().getX(), tickVolume.getSquareMin().getY(), tickVolume.getSquareMax().getZ(), QString("Z"), g_fontAxisLabel );

    int tickCount = GetParent()->ticksPerPlaneSpinbox->value();
    if (tickCount > 1)
    {
        const float tickLabelOffset = getCurrentTickLabelOffset();
        const float invTickCount = 1.0f / (float) (tickCount - 1);

        const float fXSpacing = tickVolume.getSquareWidth() * invTickCount;
        const float fYSpacing = tickVolume.getSquareHeight() * invTickCount;
        const float fZSpacing = tickVolume.getSquareDepth() * invTickCount;

        for (size_t ii = 0; ii < tickCount; ++ii)
        {
            QString strTickNumber;

            const float fXOffset = tickVolume.getSquareMin().getX() + ((float) ii * fXSpacing);
            strTickNumber.sprintf("%3.2f", fXOffset);
            renderText(fXOffset, tickVolume.getSquareMin().getY() - tickLabelOffset, tickVolume.getSquareMin().getZ(), strTickNumber);

            const float fYOffset = tickVolume.getSquareMin().getY() + ((float) ii * fYSpacing);
            strTickNumber.sprintf("%3.2f", fYOffset);
            renderText(tickVolume.getSquareMin().getX() - tickLabelOffset, fYOffset, tickVolume.getSquareMin().getZ(), strTickNumber);

            const float fZOffset = tickVolume.getSquareMin().getZ() + ((float) ii * fZSpacing);
            strTickNumber.sprintf("%3.2f", fZOffset);
            renderText(tickVolume.getSquareMin().getX() - tickLabelOffset, tickVolume.getSquareMin().getY(), fZOffset, strTickNumber);
        }
    }

    assert(checkOpenGL());
}

//void ScatterPlotWidget::updatePointColors()
//{
//    struct Lerp
//    {
//        static inline int calc(float portion, int a, int b)
//        {
//            return a + (int)((float)(b - a) * portion);
//        }
//    };
//
//    if (pointArray && pointCount)
//    {
//        // If the point color buffer is too small to contain the list of points, grow the buffer.
//        if (pointCount > pointColorCapacity)
//        {
//            delete[] pointColorArray;
//
//            size_t pointColorGrowthCapacity = (size_t)((unsigned int)pointCount * BUFFER_GROWTH_FACTOR);
//            pointColorArray = new unsigned int[ pointColorGrowthCapacity ];
//            if (pointColorArray)
//            {
//                pointColorCapacity = pointColorGrowthCapacity;
//            }
//        }
//
//        // If the color array has been successfully allocated, fill it with the point data.
//        if (pointColorArray)
//        {
//            const QPoint3F boundSize = tickVolume.getDelta();
//            for (size_t ii = 0; ii < pointCount; ++ii)
//            {
//                const size_t pointIndex = ii * 3;
//                const QPoint3F point( pointArray[ pointIndex ], pointArray[ pointIndex + 1 ], pointArray[ pointIndex + 2 ] );
//                const QPoint3F pointOffset= point - tickVolume.getMin();
//                const QPoint3F portions( pointOffset.getX() / boundSize.getX(), pointOffset.getY() / boundSize.getY(), pointOffset.getZ() / boundSize.getZ() );
//                const int red = Lerp::calc(  portions.getX(), pointMinColor.red(), pointMaxColor.red() );
//                const int green = Lerp::calc(  portions.getY(), pointMinColor.green(), pointMaxColor.green() );
//                const int blue = Lerp::calc(  portions.getZ(), pointMinColor.blue(), pointMaxColor.blue() );
//                pointColorArray[ ii ] = (red & 0xff) | ((green & 0xff) << 8) | ((blue & 0xff) << 16) | 0xff000000;
//            }
//        }
//        else
//        {
//            pointColorCapacity = 0;
//        }
//    }
//}

void ScatterPlotWidget::updateColorsByPosition(ScatterPlotFrameInput & frameInput, const QAxisAlignedBoundingBox & tickVolume)
{
    struct Lerp
    {

        static inline int calc(float portion, int a, int b)
        {
            return a + (int) ((float) (b - a) * portion);
        }
    };

    const unsigned int numberOfPoints = frameInput.numPoints;

    assert(frameInput.colorMode == SPPTCM_COLOR_BY_POSITION); // Only run this when colormode is appropriate.
    assert(numberOfPoints < MAX_POINTS); // Number of points exceeds buffer size.  Will ignore excess points in Release builds.

    if (numberOfPoints < MAX_POINTS)
    {
        const QPoint3F boundSize = tickVolume.getDelta();
        for (unsigned int ii = 0; ii < numberOfPoints; ++ii)
        {
            ColoredPoint & currentColoredPoint = frameInput.points[ ii ];
            const QPoint3F & currentPoint = currentColoredPoint.point;
            const QPoint3F pointOffset = currentPoint - tickVolume.getMin();
            const QPoint3F portions(pointOffset.getX() / boundSize.getX(), pointOffset.getY() / boundSize.getY(), pointOffset.getZ() / boundSize.getZ());
            const int red = Lerp::calc(portions.getX(), frameInput.pointPositionalMinColor.red(), frameInput.pointPositionalMaxColor.red());
            const int green = Lerp::calc(portions.getY(), frameInput.pointPositionalMinColor.green(), frameInput.pointPositionalMaxColor.green());
            const int blue = Lerp::calc(portions.getZ(), frameInput.pointPositionalMinColor.blue(), frameInput.pointPositionalMaxColor.blue());
            currentColoredPoint.perPointColor[0] = red; // = (red & 0xff) | ((green & 0xff) << 8) | ((blue & 0xff) << 16) | 0xff000000;
            currentColoredPoint.perPointColor[1] = green;
            currentColoredPoint.perPointColor[2] = blue;
        }
    }
}

void ScatterPlotWidget::updateColorsByPosition(RenderablePoint * const renderablePointBatch, unsigned int numberOfPoints, const QAxisAlignedBoundingBox & tickVolume)
{
    struct Lerp
    {

        static inline int calc(float portion, int a, int b)
        {
            return a + (int) ((float) (b - a) * portion);
        }
    };

    assert(numberOfPoints < MAX_POINTS); // Number of points exceeds buffer size.  Will ignore excess points in Release builds.

    if (numberOfPoints < MAX_POINTS)
    {
        const QPoint3F boundSize = tickVolume.getDelta();
        for (unsigned int ii = 0; ii < numberOfPoints; ++ii)
        {
            RenderablePoint & currentColoredPoint = renderablePointBatch[ ii ];
            const QPoint3F & currentPoint = currentColoredPoint.point;
            const QPoint3F pointOffset = currentPoint - tickVolume.getMin();
            const QPoint3F portions(pointOffset.getX() / boundSize.getX(), pointOffset.getY() / boundSize.getY(), pointOffset.getZ() / boundSize.getZ());
            const int red = Lerp::calc(portions.getX(), currentColoredPoint.frameInput->pointPositionalMinColor.red(), currentColoredPoint.frameInput->pointPositionalMaxColor.red());
            const int green = Lerp::calc(portions.getY(), currentColoredPoint.frameInput->pointPositionalMinColor.green(), currentColoredPoint.frameInput->pointPositionalMaxColor.green());
            const int blue = Lerp::calc(portions.getZ(), currentColoredPoint.frameInput->pointPositionalMinColor.blue(), currentColoredPoint.frameInput->pointPositionalMaxColor.blue());
            currentColoredPoint.pointColor[0] = red; // = (red & 0xff) | ((green & 0xff) << 8) | ((blue & 0xff) << 16) | 0xff000000;
            currentColoredPoint.pointColor[1] = green;
            currentColoredPoint.pointColor[2] = blue;
        }
    }
}

void ScatterPlotWidget::calculateInitialDistance(float minCameraDistance)
{
    cameraDefaultDistance = 0.0f;
    for (int ii = 1;; ++ii)
    {
        cameraDefaultDistance = logToDistance(ii);
        cameraCurrentDistance = cameraDefaultDistance;
        if (cameraDefaultDistance >= minCameraDistance)
        {
            cameraLogDistance = ii;
            break;
        }
    }
}

float ScatterPlotWidget::logToDistance(int intLog)
{
    float scaler = (float) intLog * 0.01f;
    return pow(10.0f, scaler) * 0.0001f;
}

bool ScatterPlotWidget::setPlotPoints(float * const pointArraySrc, size_t pointCountSrc, bool copy)
{
    assert(pointArraySrc);
    assert(pointCountSrc);

    bool success = false;

    if (copy)
    {
        const size_t pointArrayLength = pointCountSrc * 3;
        const size_t pointArrayMemSize = pointArrayLength * sizeof ( float);

        // If the incoming point list is larger than the current point list buffer, grow the point buffer
        if (pointCountSrc > pointCapacity)
        {
            delete[] pointArray;

            size_t pointGrowthCapacity = (size_t) ((float) pointCountSrc * BUFFER_GROWTH_FACTOR);
            pointArray = new float[ pointGrowthCapacity * 3 ];
            if (pointArray)
            {
                pointCapacity = pointGrowthCapacity;
            }
        }

        if (pointArray)
        {
            ::memcpy(pointArray, pointArraySrc, pointArrayMemSize);
            pointCount = pointCountSrc;
        }
        else
        {
            pointCount = 0;
        }
    }
    else
    {
        // If there is a local copy of the point array, delete it.  We are using the provided point array directly [faster].
        if (pointCapacity && pointArray)
        {
            delete[] pointArray;
            pointCapacity = 0;
        }

        pointArray = pointArraySrc;
        pointCount = pointCountSrc;
    }

    if (pointArray)
    {
        //updatePointColors();

        success = true;
    }

    return success;
}

bool ScatterPlotWidget::createTexture(const QImage & qImage, GLuint & textureId)
{
    bool bSuccess = false;

    makeCurrent();
    if (isValid())
    {
        GLuint oldTextureId = textureId;
        GLuint newTextureId;

        if (!qImage.isNull())
        {
            newTextureId = bindTexture(qImage, GL_TEXTURE_2D, GL_RGBA);
            bSuccess = newTextureId != 0;
            if (bSuccess)
            {
                if (oldTextureId)
                    deleteTexture(oldTextureId);
                textureId = newTextureId;
            }
        }
    }

    return bSuccess;
}

void ScatterPlotWidget::freeTexture(GLuint & textureId)
{
    if (textureId)
    {
        makeCurrent();
        if (isValid())
        {
            deleteTexture(textureId);

            assert(checkOpenGL());

            textureId = 0;
        }
    }
}

void ScatterPlotWidget::Save(XMLSaver &xmlSaver)
{
    // crashes, don't do this
    //	QByteArray geo_data;
    //	geo_data = saveGeometry();
    //	XML_SAVER_SAVE_ELEMENT(xmlSaver, geo_data);

	xmlSaver.SaveElement("backgroundTextureFilename", backgroundTextureFilename);
	xmlSaver.SaveElement("showPlanesCheckBox", GetParent()->showPlanesCheckBox);
	xmlSaver.SaveElement("ticksPerPlaneSpinbox", GetParent()->ticksPerPlaneSpinbox);
	xmlSaver.SaveElement("userDefinedCameraHeightSpinbox", GetParent()->userDefinedCameraHeightSpinbox);
	xmlSaver.SaveElement("userDefinedCameraAngleSpinbox", GetParent()->userDefinedCameraAngleSpinbox);
	xmlSaver.SaveElement("xAxisRadioButton", GetParent()->xAxisRadioButton);
	xmlSaver.SaveElement("yAxisRadioButton", GetParent()->yAxisRadioButton);
	xmlSaver.SaveElement("zAxisRadioButton", GetParent()->zAxisRadioButton);
	xmlSaver.SaveElement("perspectiveRadioButton", GetParent()->perspectiveRadioButton);
	xmlSaver.SaveElement("userDefinedRadioButton", GetParent()->userDefinedRadioButton);
	xmlSaver.SaveElement("cameraCurrentDistance", cameraCurrentDistance);
	xmlSaver.SaveElement("cameraCurrentHeight", cameraCurrentHeight);
	xmlSaver.SaveElement("cameraCurrentAngle", cameraCurrentAngle);
}

void ScatterPlotWidget::Load(XMLSaver &xmlSaver)
{
    // crashes, don't do this
    //	QByteArray geo_data;
    //	XML_SAVER_LOAD_ELEMENT(xmlSaver, geo_data, false);
    //	restoreGeometry(geo_data);

    QString backgroundTextureFilename;
	xmlSaver.LoadElement("backgroundTextureFilename",backgroundTextureFilename,"");
	if (backgroundTextureFilename.isEmpty() == false)
        LoadBackground(backgroundTextureFilename);
    else
        ClearBackground();
	
	xmlSaver.LoadElement("showPlanesCheckBox", GetParent()->showPlanesCheckBox, true);
	xmlSaver.LoadElement("ticksPerPlaneSpinbox", GetParent()->ticksPerPlaneSpinbox, QString::number(GetParent()->ticksPerPlaneSpinbox->value()));
	xmlSaver.LoadElement("userDefinedCameraHeightSpinbox", GetParent()->userDefinedCameraHeightSpinbox, QString::number(GetParent()->userDefinedCameraHeightSpinbox->value()));
	xmlSaver.LoadElement("userDefinedCameraAngleSpinbox", GetParent()->userDefinedCameraAngleSpinbox, QString::number(GetParent()->userDefinedCameraAngleSpinbox->value()));
	xmlSaver.LoadElement("xAxisRadioButton", GetParent()->xAxisRadioButton, false);
	xmlSaver.LoadElement("yAxisRadioButton", GetParent()->yAxisRadioButton, false);
	xmlSaver.LoadElement("zAxisRadioButton", GetParent()->zAxisRadioButton, false);
	xmlSaver.LoadElement("perspectiveRadioButton", GetParent()->perspectiveRadioButton, true);
	xmlSaver.LoadElement("userDefinedRadioButton", GetParent()->userDefinedRadioButton, false);
	xmlSaver.LoadElement("cameraCurrentDistance", cameraCurrentDistance, QString::number(cameraCurrentDistance));
	xmlSaver.LoadElement("cameraCurrentHeight", cameraCurrentHeight, QString::number(cameraCurrentHeight));
	xmlSaver.LoadElement("cameraCurrentAngle", cameraCurrentAngle, QString::number(cameraCurrentAngle));

//     if (GetParent()->xAxisRadioButton->isChecked())
//         on_xAxisRadioButton_released();
//     if (GetParent()->yAxisRadioButton->isChecked())
//         on_yAxisRadioButton_released();
// 	if (GetParent()->zAxisRadioButton->isChecked())
// 		on_zAxisRadioButton_released();
// 	if (GetParent()->perspectiveRadioButton->isChecked())
// 		on_perspectiveRadioButton_released();
// 	if (GetParent()->userDefinedRadioButton->isChecked())
// 		on_userDefinedRadioButton_released();
}